#!/usr/bin/env python

# stdlib imports
import os.path
import sys
import time
import datetime

# third party imports
import matplotlib.pyplot as plt
import numpy as np
import pytest
from esi_utils_rupture.origin import Origin
from esi_utils_rupture.quad_rupture import QuadRupture
from mpl_toolkits.mplot3d import axes3d
from openquake.hazardlib.geo.geodetic import azimuth

# local imports
from esi_shakelib.plotting.plotrupture import map_rupture, plot_rupture_wire3d

homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, ".."))
sys.path.insert(0, shakedir)

MAX_DEPTH = 70
DIP = 17


def test_plot_rupture(interactive=False):
    xp0 = np.array([-90.898000])
    xp1 = np.array([-91.308000])
    yp0 = np.array([12.584000])
    yp1 = np.array([12.832000])
    zp = [0.0]
    strike = azimuth(yp0[0], xp0[0], yp1[0], xp1[0])
    origin = Origin(
        {
            "lat": 0.0,
            "lon": 0.0,
            "depth": 0.0,
            "mag": 5.5,
            "id": "",
            "netid": "abcd",
            "network": "",
            "locstring": "",
            "time": datetime.datetime.fromtimestamp(time.time(), datetime.UTC),
        }
    )
    interface_width = MAX_DEPTH / np.sin(np.radians(DIP))
    widths = np.ones(xp0.shape) * interface_width
    dips = np.ones(xp0.shape) * DIP
    strike = [strike]
    rupture = QuadRupture.fromTrace(
        xp0, yp0, xp1, yp1, zp, widths, dips, origin, strike=strike
    )
    plot_rupture_wire3d(rupture)
    if interactive:
        fname = os.path.join(os.path.expanduser("~"), "rupture_wire_plot.png")
        plt.savefig(fname)
        print(f"Wire 3D plot saved to {fname}.  Delete this file if you wish.")

    # Need to get tests to check exception for if an axis is handed off
    fig = plt.figure()
    # ax = fig.add_subplot(111, projection="3d")
    ax = axes3d.Axes3D(fig)
    plot_rupture_wire3d(rupture, ax)

    # And raise the exception if it is not a 3d axis
    with pytest.raises(TypeError):
        ax = fig.add_subplot(111)
        plot_rupture_wire3d(rupture, ax)


def test_map_rupture(interactive=False):
    xp0 = np.array([-90.898000])
    xp1 = np.array([-91.308000])
    yp0 = np.array([12.584000])
    yp1 = np.array([12.832000])
    zp = [0.0]
    strike = azimuth(yp0[0], xp0[0], yp1[0], xp1[0])
    origin = Origin(
        {
            "lat": 0.0,
            "lon": 0.0,
            "depth": 0.0,
            "mag": 5.5,
            "id": "",
            "netid": "abcd",
            "network": "",
            "locstring": "",
            "time": datetime.datetime.fromtimestamp(time.time(), datetime.UTC),
        }
    )
    interface_width = MAX_DEPTH / np.sin(np.radians(DIP))
    widths = np.ones(xp0.shape) * interface_width
    dips = np.ones(xp0.shape) * DIP
    strike = [strike]
    rupture = QuadRupture.fromTrace(
        xp0, yp0, xp1, yp1, zp, widths, dips, origin, strike=strike
    )
    map_rupture(rupture)
    if interactive:
        fname = os.path.join(os.path.expanduser("~"), "rupture_map.png")
        plt.savefig(fname)
        print(
            f"Rupture map plot saved to {fname}.  Delete this file if you wish."
        )


if __name__ == "__main__":
    test_plot_rupture(interactive=True)
    test_map_rupture(interactive=True)



#!/usr/bin/env python

# stdlib imports

# third party imports
from esi_utils_io.smcontainers import ShakeMapContainerBase
from esi_utils_rupture.factory import (
    get_rupture,
    rupture_from_dict,
    rupture_from_dict_and_origin,
)
from esi_utils_rupture.origin import Origin

# local imports
from esi_shakelib.station import StationList


class ShakeMapInputContainer(ShakeMapContainerBase):
    """
    HDF container for Shakemap input data.

    This class provides methods for getting and setting information on:
     - configuration
     - event (lat,lon,depth,etc.)
     - rupture
     - station data (can also be updated)
     - version history

    """

    @classmethod
    def createFromInput(
        cls,
        filename,
        config,
        eventfile,
        version_history,
        rupturefile=None,
        sourcefile=None,
        momentfile=None,
        datafiles=None,
    ):
        """
        Instantiate an InputContainer from ShakeMap input data.

        Args:
            filename (str): Path to HDF5 file that will be created to
                encapsulate all input data.
            config (dict): Dictionary containing all configuration information
                necessary for ShakeMap ground motion and other calculations.
            eventfile (str): Path to ShakeMap event.xml file.
            rupturefile (str): Path to ShakeMap rupture text or JSON file.
            sourcefile (str): Path to ShakeMap source.txt file.
            momentfile (str): Path to ShakeMap moment.xml file.
            datafiles (list): List of ShakeMap data (DYFI, strong motion)
                files.
            version_history (dict): Dictionary containing version history.

        Returns:
            InputContainer: Instance of InputContainer.
        """
        container = cls.create(filename)
        container.setConfig(config)

        # create an origin from the event file
        origin = Origin.fromFile(
            eventfile, sourcefile=sourcefile, momentfile=momentfile
        )

        # create a rupture object from the origin and the rupture file
        # (if present).
        rupture = get_rupture(origin, file=rupturefile)

        # store the rupture object in the container
        container.setRupture(rupture)

        if datafiles is not None and len(datafiles) > 0:
            container.setStationData(datafiles)

        container.setVersionHistory(version_history)
        return container

    def setRupture(self, rupture):
        """
        Store Rupture object in container.

        Args:
            rupture (dict or Rupture): Rupture object (Point,Quad, or Edge)
                or dictionary representation of same.
        Raises:
            TypeError: If input object or dictionary does not
                represent a Rupture object.
        """
        if "rupture" in self.getDictionaries():
            self.dropDictionary([], "rupture")
        if isinstance(rupture, dict):
            try:
                rupture_from_dict(rupture)
            except Exception:
                fmt = "Input dict does not represent a rupture object."
                raise TypeError(fmt)
            rupdict = rupture
        else:
            rupdict = rupture._geojson
        self.setRuptureDict(rupdict)

    def getRuptureObject(self):
        """
        Retrieve Rupture object from container.

        Returns:
            Rupture: Instance of (one of) a Point/Quad/EdgeRupture class.
        Raises:
            AttributeError: If rupture object has not been set in
                the container.
        """
        if "rupture" not in self.getDictionaries():
            raise AttributeError("Rupture dictionary not set in container.")
        rupture_dict = self.getRuptureDict()
        rupture = rupture_from_dict(rupture_dict)
        return rupture

    def setStrecJson(self, strec):
        """
        Store STREC JSON string in container

        Args:
            strec (JSON): JSON string of STREC output.
        Raises:
            TypeError: If input object is not a string.
        """
        if "strec" in self.getStrings():
            self.dropString([], "strec")
        if not isinstance(strec, str):
            fmt = "Input object is not a string."
            raise TypeError(fmt)
        self.setString([], "strec", strec)

    def getStrecJson(self):
        """
        Retrieve STREC JSON from container.

        Returns:
            String: String of STREC JSON.
        Raises:
            AttributeError: If STREC JSON string has not been set in
                the container.
        """
        if "strec" not in self.getStrings():
            raise AttributeError("strec object not set in container.")
        strec = self.getString([], "strec")
        return strec

    def setStationList(self, stationlist):
        """
        Store StationList object in container.

        Args:
            stationlist (StationList): StationList object.
        Raises:
            TypeError: If input object or dictionary is not a StationList
                object.
        """
        if "stations" in self.getStrings():
            self.dropString([], "stations")
        if not isinstance(stationlist, StationList):
            fmt = "Input object is not a StationList."
            raise TypeError(fmt)
        sql_string = stationlist.dumpToSQL()
        self.setString([], "stations", sql_string)

    def getStationList(self):
        """
        Retrieve StationList object from container.

        Returns:
            StationList: StationList object.
        Raises:
            AttributeError: If stationlist object has not been set in
                the container.
        """
        if "stations" not in self.getStrings():
            raise AttributeError("StationList object not set in container.")
        sql_string = self.getString([], "stations")
        stationlist = StationList.loadFromSQL(sql_string)
        return stationlist

    def setStationData(self, datafiles):
        """
        Insert observed ground motion data into the container.

        Args:
          datafiles (str): Path to XML- or JSON-formatted files containing
              ground motion observations, (macroseismic or instrumented).

        """
        station = StationList.loadFromFiles(
            datafiles, min_nresp=self.getConfig()["data"]["min_nresp"]
        )
        self.setStationList(station)

    def addStationData(self, datafiles):
        """
        Add observed ground motion data into the container.

        Args:
            datafiles (sequence): Sequence of paths to XML- and/or
                JSON-formatted files containing ground motion observations,
                (macroseismic or instrumented).
        """
        if len(datafiles) == 0:
            return
        try:
            station = self.getStationList()
            station.addData(datafiles, self.getConfig()["data"]["min_nresp"])
        except AttributeError:
            station = StationList.loadFromFiles(
                datafiles, min_nresp=self.getConfig()["data"]["min_nresp"]
            )
        self.setStationList(station)

    def updateRupture(self, eventxml=None, rupturefile=None):
        """Update rupture/origin information in container.

        Args:
            eventxml (str): Path to event.xml file.
            rupturefile (str): Path to rupture file (JSON or txt format).
        """
        if eventxml is None and rupturefile is None:
            return

        # the container is guaranteed to have at least a Point rupture
        # and the origin.
        rupture = self.getRuptureObject()
        origin = rupture.getOrigin()

        if eventxml is not None:
            origin = Origin.fromFile(eventxml)
            if rupturefile is not None:
                rupture = get_rupture(origin, file=rupturefile)
            else:
                rupture_dict = rupture._geojson
                rupture = rupture_from_dict_and_origin(rupture_dict, origin)
        else:  # no event.xml file, but we do have a rupture file
            rupture = get_rupture(origin, file=rupturefile)

        self.setRupture(rupture)



"""
A GMPE that returns a constant everywhere. Useful for testing, but
nothing else.
"""

import numpy as np

from openquake.hazardlib.gsim.base import GMPE, CoeffsTable
from openquake.hazardlib import const
from openquake.hazardlib.imt import PGA, PGV, SA


class NullGMPE(GMPE):
    """
    This is a GMPE for testing. It returns the mean and stddevs
    specified in the constructor.
    """

    DEFINED_FOR_TECTONIC_REGION_TYPE = const.TRT.ACTIVE_SHALLOW_CRUST
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = set([PGA, PGV, SA])
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = const.IMC.GREATER_OF_TWO_HORIZONTAL
    DEFINED_FOR_STANDARD_DEVIATION_TYPES = set(
        [const.StdDev.TOTAL, const.StdDev.INTER_EVENT, const.StdDev.INTRA_EVENT]
    )
    REQUIRES_SITES_PARAMETERS = set(("vs30",))
    REQUIRES_RUPTURE_PARAMETERS = set(())
    REQUIRES_DISTANCES = set(("rjb",))

    def __init__(self, mean=0, phi=0.8, tau=0.6):
        """
        The default constructor takes three named arguments:

        Args:
            mean (float): the mean value returned by the GMPE (default=0).
                This value is returned for all locations, regardles of
                the IMT or the contents of sites, rupture and distance
                contexts.
            phi (float): the within-event standard deviation returned by the
                 GMPE (default=0.8)
            tau (float): the between-event standard deviation returned by the
                 GMPE (default=0.6)

        The total standard deviation returned will be ``sqrt(phi^2 + tau^2)``.
        """
        self.mean = mean
        self.phi = phi
        self.tau = tau
        self.sigma = np.sqrt(phi**2 + tau**2)

    def get_mean_and_stddevs(self, sites, rup, dists, imt, stddev_types):
        """
        Implements the OpenQuake GroundShakingIntensityModel
        get_mean_and_stddevs interface. See superclass
        `method <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#openquake.hazardlib.gsim.base.GroundShakingIntensityModel.get_mean_and_stddevs>`__.

        Returns a constant values for all locations specified in
        the dists.rbj array, regardless of the contents of that array or
        any of the other contexts. The imt is also ignored.
        """  # noqa

        if isinstance(imt, list):
            nlist = len(imt)
        else:
            nlist = 1
        N = np.size(dists.rjb)
        mean = np.zeros((nlist, N))
        stddevs = []
        for i in range(nlist):
            for stddev_type in stddev_types:
                if stddev_type == const.StdDev.TOTAL:
                    stddevs.append(
                        np.full_like(dists.rjb, np.sqrt(self.phi**2 + self.tau**2))
                    )
                elif stddev_type == const.StdDev.INTRA_EVENT:
                    stddevs.append(np.full_like(dists.rjb, self.phi))
                elif stddev_type == const.StdDev.INTER_EVENT:
                    stddevs.append(np.full_like(dists.rjb, self.tau))

        return mean, stddevs

    #
    # Dummy COEFFS table so MultiGMPE won't complain
    #
    COEFFS = CoeffsTable(
        sa_damping=5,
        table="""\
IMT     c1
pga    0.
pgv    0.
0.01   0.
10     0.
""",
    )



#!/usr/bin/env python

# stdlib imports
import copy
import logging
from importlib import import_module

# third party imports
import numpy as np
from openquake.hazardlib import const
from openquake.hazardlib.gsim.base import GMPE
from openquake.hazardlib.gsim.boore_2014 import BooreEtAl2014
from openquake.hazardlib.gsim.campbell_bozorgnia_2014 import (
    CampbellBozorgnia2014,
)
from openquake.hazardlib.imt import PGA, PGV, SA
from openquake.hazardlib.valid import gsim

# local imports
from esi_shakelib.utils.contexts import RuptureContext
from esi_shakelib.utils.gmpe_coeffs import get_gmpe_sa_periods
from esi_shakelib.conversions.imc.boore_kishida_2017 import BooreKishida2017
from esi_shakelib.conversions.imt.abrahamson_bhasin_2020 import (
    AbrahamsonBhasin2020,
)
from esi_shakelib.multiutils import gmpe_gmas
from esi_shakelib.sites import addDepthParameters

# These are imported for their side effects...
import esi_shakelib.gmpe.atkinson_macias_2009
import esi_shakelib.gmpe.zhao_2006
import esi_shakelib.gmpe.kuehn_2020
import esi_shakelib.gmpe.ak_bias_gmms
import esi_shakelib.gmpe.nga_east


def set_sites_depth_parameters(sites, gmpe):
    """
    Need to select the appropriate z1pt0 value for different GMPEs.
    Note that these are required site parameters, so even though
    OQ has these equations built into the class in most cases.
    I have submitted an issue to OQ requesting subclasses of these
    methods that do not require the depth parameters in the
    SitesContext to make this easier.

    Args:
        sites:1 An OQ sites context.
        gmpe: An OQ GMPE instance.

    Returns:
        An OQ sites context with the depth parameters set for the
        requested GMPE.
    """
    if gmpe == "[MultiGMPE]":
        return sites

    addDepthParameters(sites)

    if gmpe in (
        "[AbrahamsonEtAl2014]",
        '[AbrahamsonEtAl2014]\nregion = "TWN"',
        '[AbrahamsonEtAl2014]\nregion = "CHN"',
    ):
        sites.z1pt0 = sites.z1pt0_ask14_cal
    if gmpe == '[AbrahamsonEtAl2014]\nregion = "JPN"':
        sites.z1pt0 = sites.z1pt0_ask14_jpn
    if gmpe == "[ChiouYoungs2014]" or isinstance(gmpe, BooreEtAl2014):
        sites.z1pt0 = sites.z1pt0_cy14_cal

    gmpe_name = f"{gmpe}"
    if "KuehnEtAl202" in gmpe_name:
        if "Japan" in gmpe_name:
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
    if "AbrahamsonGulerce2020S" in gmpe_name:
        if "Japan" in gmpe_name:
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
    if "ZhaoEtAl2006S" in gmpe_name:
        if "Japan" in gmpe_name:
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
    if "AtkinsonMacias2009" in gmpe_name:
        if "Japan" in gmpe_name:
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
    if "KuehnEtAl2020S" in gmpe_name:
        # This is just a WAG at what this GMPE wants
        if "Japan" in gmpe_name:
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
        sites.z1pt0 = sites.z1pt0_cy14_cal

    if "KothaEtAl2020ESHM20" in gmpe_name:
        sites.region = np.zeros_like(sites.vs30)

    if isinstance(gmpe, CampbellBozorgnia2014):
        if (
            gmpe == "[CampbellBozorgnia2014JapanSite]"
            or gmpe == "[CampbellBozorgnia2014HighQJapanSite]"
            or gmpe == "[CampbellBozorgnia2014LowQJapanSite]"
        ):
            sites.z2pt5 = sites.z2pt5_cb14_jpn
        else:
            sites.z2pt5 = sites.z2pt5_cb14_cal
    if (
        gmpe == "[ChiouYoungs2008]"
        or gmpe == "[Bradley2013]"
        or gmpe == "[Bradley2013Volc]"
    ):
        sites.z1pt0 = sites.z1pt0_cy08
    if gmpe == "[CampbellBozorgnia2008]":
        sites.z2pt5 = sites.z2pt5_cb07
    if gmpe == "[AbrahamsonSilva2008]":
        sites.z1pt0 = gmpe._compute_median_z1pt0(sites.vs30)

    return sites


def stuff_context(sites, rup, dists):
    """
    Function to fill a rupture context with the contents of all of the
    other contexts.

    Args:
        sites (SiteCollection): A SiteCollection object.

        rup (RuptureContext): A RuptureContext object.

        dists (DistanceContext): A DistanceContext object.

    Returns:
        RuptureContext: A new RuptureContext whose attributes are all of
        the elements of the three inputs.
    """
    if (
        isinstance(sites, np.recarray)
        and isinstance(rup, np.recarray)
        and isinstance(dists, np.recarray)
        and np.all(sites == rup)
        and np.all(rup == dists)
    ):
        # We're already a rec array
        return rup

    shape = None
    ctx = RuptureContext()
    for name, val in vars(sites).items():
        if name.startswith("__"):
            continue
        # We don't want the sites lons and lats, we want the dists ones
        if name in ["lons", "lats"]:
            continue
        if shape is None:
            shape = val.shape
        setattr(ctx, name, val)
    for name, val in vars(rup).items():
        if name.startswith("__"):
            continue
        if not isinstance(val, np.ndarray):
            val = np.full(shape, val)
        setattr(ctx, name, val)
    for name, val in vars(dists).items():
        if name.startswith("__"):
            continue
        setattr(ctx, name, val)

    ctx.occurrence_rate = np.full(shape, 1e-5)

    cols = []
    arrs = []
    for key, val in vars(ctx).items():
        if key.startswith("__"):
            continue
        cols.append(key)
        arrs.append(val)
    cx = np.rec.fromarrays(arrs, names=cols)
    return cx


def get_gmpe_from_name(name, conf):

    # Only import the NullGMPE when we're testing
    # We'll want to import any other GMPEs we add at the top of this module
    # so that gsim() picks them up; anything in OQ is already included
    if name == "NullGMPE":
        _ = import_module(conf["gmpe_modules"][name][1])
    return gsim(name)


class MultiGMPE(GMPE):
    """
    Implements a GMPE that is the combination of multiple GMPEs.

    """

    DEFINED_FOR_TECTONIC_REGION_TYPE = None
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = None
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = None
    DEFINED_FOR_STANDARD_DEVIATION_TYPES = set([const.StdDev.TOTAL])
    REQUIRES_SITES_PARAMETERS = None
    REQUIRES_RUPTURE_PARAMETERS = None
    REQUIRES_DISTANCES = None

    def get_mean_and_stddevs(self, sites, rup, dists, imt, stddev_types):
        """
        See superclass `method <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#openquake.hazardlib.gsim.base.GroundShakingIntensityModel.get_mean_and_stddevs>`__.

        """  # noqa

        for key, val in rup.__dict__.items():
            if not isinstance(val, np.ndarray):
                setattr(rup, key, np.full_like(sites.vs30, val))

        sd_avail = self.DEFINED_FOR_STANDARD_DEVIATION_TYPES
        if not sd_avail.issuperset(set(stddev_types)):
            raise TypeError("Requested an unavailable stddev_type.")

        if not isinstance(imt, list):
            imt = [imt]
        # Evaluate MultiGMPE:
        lnmu, lnsd = self.__get_mean_and_stddevs__(
            sites, rup, dists, imt, stddev_types
        )

        return lnmu, lnsd

    def __get_mean_and_stddevs__(self, sites, rup, dists, imt, stddev_types):

        nsd = len(stddev_types)
        # ---------------------------------------------------------------------
        # Sort out which set of weights to use
        # ---------------------------------------------------------------------
        wts = self.WEIGHTS

        nimt = len(imt)
        # ---------------------------------------------------------------------
        # This is the array to hold the weighted combination of the GMPEs
        # ---------------------------------------------------------------------
        lnmu = np.zeros((nimt, np.size(sites.vs30)))
        # ---------------------------------------------------------------------
        # Hold on to the individual means and stddevs so we can compute the
        # combined stddev
        # ---------------------------------------------------------------------
        lnmu_list = [[] for _ in range(nimt)]
        lnsd_list = [[] for _ in range(nimt)]

        for i, gmpe in enumerate(self.GMPES):
            # -----------------------------------------------------------------
            # Loop over GMPE list
            # -----------------------------------------------------------------

            set_sites_depth_parameters(sites, gmpe)

            # -----------------------------------------------------------------
            # Select the IMT
            # -----------------------------------------------------------------

            gmpe_imts = [
                imt.__name__
                for imt in list(gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES)
            ]

            timt = copy.copy(imt)
            if (
                not isinstance(gmpe, MultiGMPE)
                and PGV() in timt
                and ("PGV" not in gmpe_imts)
            ):
                ab2020 = AbrahamsonBhasin2020(rup.mag)
                pgv_ix = timt.index(PGV())
                timt[pgv_ix] = SA(ab2020.getTref())

            # -----------------------------------------------------------------
            # Grab GMPE_LIMITS in gmpe instance for later as the multigmpe
            # nests downward.
            # -----------------------------------------------------------------
            if hasattr(self, "GMPE_LIMITS"):
                # Remember that GMPE_LIMITS is only present if it is getting
                # loaded from a config... we could change this eventually.
                gmpe.GMPE_LIMITS = self.GMPE_LIMITS

            # -----------------------------------------------------------------
            # Apply GMPE_LIMITS if applicable
            # -----------------------------------------------------------------
            if hasattr(gmpe, "GMPE_LIMITS"):
                gmpes_with_limits = list(gmpe.GMPE_LIMITS.keys())
                gmpe_class_str = str(gmpe).replace("[", "").replace("]", "")
                if gmpe_class_str in gmpes_with_limits:
                    limit_dict = gmpe.GMPE_LIMITS[gmpe_class_str]
                    for k, v in limit_dict.items():
                        if k == "vs30":
                            vs30min = float(v[0])
                            vs30max = float(v[1])
                            sites.vs30 = np.clip(sites.vs30, vs30min, vs30max)
                            addDepthParameters(sites)
                        elif k == "ztor":
                            ztor_min = float(v[0])
                            ztor_max = float(v[1])
                            rup.ztor = np.clip(rup.ztor, ztor_min, ztor_max)

            # -----------------------------------------------------------------
            # Evaluate
            # -----------------------------------------------------------------
            if not isinstance(gmpe, MultiGMPE):
                ctx = stuff_context(sites, rup, dists)
                lmean, lsd = gmpe_gmas(gmpe, ctx, timt, stddev_types)
            else:
                lmean, lsd = gmpe.get_mean_and_stddevs(
                    sites, rup, dists, timt, stddev_types
                )

            if not isinstance(gmpe, MultiGMPE):
                # -------------------------------------------------------------
                # If IMT is PGV and PGV is not given by the GMPE, then
                # convert from the appropriate PSA
                # -------------------------------------------------------------
                if (PGV() in imt) and ("PGV" not in gmpe_imts):
                    tmean, tsd = ab2020.getPGVandSTDDEVS(
                        lmean[pgv_ix],
                        lsd[pgv_ix * nsd : (pgv_ix + 1) * nsd],
                        stddev_types,
                        ctx.rrup,
                        ctx.vs30,
                    )
                    lmean[pgv_ix] = tmean
                    for ix, _ in enumerate(stddev_types):
                        lsd[pgv_ix * nsd + ix] = tsd[ix]

                # -------------------------------------------------------------
                # Convertions due to component definition
                # -------------------------------------------------------------
                imc_in = gmpe.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT
                imc_out = self.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT
                if imc_in != imc_out:
                    bk17 = BooreKishida2017(imc_in, imc_out)
                    for kk, kimt in enumerate(timt):
                        if kimt.string == "MMI":
                            continue
                        lmean[kk] = bk17.convertAmps(
                            kimt, lmean[kk], dists.rrup, rup.mag
                        )
                        #
                        # The extra sigma from the component conversion
                        # appears to apply to the total sigma, so the
                        # question arises as to how to apportion it between
                        # the intra- and inter-event sigma. Here we assume
                        # it all enters as intra-event sigma.
                        #
                        for j, stddev_type in enumerate(stddev_types):
                            if stddev_type == const.StdDev.INTER_EVENT:
                                continue
                            lsd[kk * nsd + j] = bk17.convertSigmas(
                                kimt, lsd[kk * nsd + j]
                            )

            # End: if GMPE is not MultiGMPE

            #
            # At this point lsd will have nimt * len(stddev_types) entries, the
            #

            # -----------------------------------------------------------------
            # Compute weighted mean and collect the elements to compute sd
            # -----------------------------------------------------------------

            for kk, kimt in enumerate(timt):
                lnmu[kk] = lnmu[kk] + wts[i] * lmean[kk]
                lnmu_list[kk].append(lmean[kk])
                lnsd_list[kk] = lnsd_list[kk] + lsd[kk * nsd : (kk + 1) * nsd]

        # -----------------------------------------------------------------
        # The mean is a weighted sum of random variables, so the stddev
        # is the weighted sum of of their covariances (effectively). See:
        # https://en.wikipedia.org/wiki/Variance#Weighted_sum_of_variables
        # for an explanation. Also see:
        # http://usgs.github.io/shakemap/manual4_0/tg_processing.html#ground-motion-prediction
        # for a discussion on the way this is implemented here.
        # -------------------------------------------------------------- # noqa

        nwts = len(wts)
        npwts = np.array(wts).reshape((1, -1))
        lnsd_new = []
        for kk, kimt in enumerate(timt):
            nsites = len(lnmu[kk])
            # Find the correlation coefficients among the gmpes; if there are
            # fewer than 10 points, just use an approximation (noting that the
            # correlation among GMPEs tends to be quite high).
            if nsites < 10:
                cc = np.full((nwts, nwts), 0.95)
                np.fill_diagonal(cc, 1.0)
            else:
                np.seterr(divide="ignore", invalid="ignore")
                cc = np.reshape(np.corrcoef(lnmu_list[kk]), (nwts, nwts))
                np.seterr(divide="warn", invalid="warn")
                cc[np.isnan(cc)] = 1.0

            # Multiply the correlation coefficients by the weights matrix
            # (this is cheaper than multiplying all of elements of each
            # stddev array by their weights since we have to multiply
            # everything by the correlation coefficient matrix anyway))
            cc = ((npwts * npwts.T) * cc).reshape((nwts, nwts, 1))
            for i in range(nsd):
                sdlist = []
                for j in range(nwts):
                    sdlist.append(
                        lnsd_list[kk][j * nsd + i].reshape((1, 1, -1))
                    )
                sdstack = np.hstack(sdlist)
                wcov = (sdstack * np.transpose(sdstack, axes=(1, 0, 2))) * cc
                # This sums the weighted covariance as each point in the output
                lnsd_new.append(np.sqrt(wcov.sum((0, 1))))

        return (lnmu, lnsd_new)

    @classmethod
    def __from_config__(cls, conf, filter_imt=None):
        """
        Construct a MultiGMPE from a config file.

        Args:
            conf (dict): Dictionary of config options.
            filter_imt (IMT): An optional IMT to filter/reweight the GMPE list.

        Returns:
            MultiGMPE object.

        """
        imc = getattr(const.IMC, conf["interp"]["component"])
        selected_gmpe = conf["modeling"]["gmpe"]

        logging.debug("selected_gmpe: %s", selected_gmpe)
        logging.debug("IMC: %s", imc)

        # ---------------------------------------------------------------------
        # Allow for selected_gmpe to be found in either conf['gmpe_sets'] or
        # conf['gmpe_modules'], if it is a GMPE set, then all entries must be
        # either a GMPE or a GMPE set (cannot have a GMPE set that is a mix of
        # GMPEs and GMPE sets).
        # ---------------------------------------------------------------------

        if selected_gmpe in conf["gmpe_sets"].keys():
            selected_gmpe_sets = conf["gmpe_sets"][selected_gmpe]["gmpes"]
            gmpe_set_weights = [
                float(w) for w in conf["gmpe_sets"][selected_gmpe]["weights"]
            ]
            logging.debug("selected_gmpe_sets: %s", selected_gmpe_sets)
            logging.debug("gmpe_set_weights: %s", gmpe_set_weights)

            # -----------------------------------------------------------------
            # If it is a GMPE set, does it contain GMPEs or GMPE sets?
            # -----------------------------------------------------------------

            set_of_gmpes = all(
                s in conf["gmpe_modules"] for s in selected_gmpe_sets
            )
            set_of_sets = all(
                s in conf["gmpe_sets"] for s in selected_gmpe_sets
            )

            if set_of_sets is True:
                mgmpes = []
                mweights = []
                for ix, s in enumerate(selected_gmpe_sets):
                    mgmpe = cls.__multigmpe_from_gmpe_set__(
                        conf, s, filter_imt=filter_imt
                    )
                    if mgmpe is None:
                        continue
                    mgmpes.append(mgmpe)
                    mweights.append(gmpe_set_weights[ix])
                if len(mgmpes) == 0:
                    return None
                mweights = np.array(mweights)
                mweights = mweights / np.sum(mweights)
                out = MultiGMPE.__from_list__(mgmpes, mweights, imc=imc)
            elif set_of_gmpes is True:
                out = cls.__multigmpe_from_gmpe_set__(
                    conf, selected_gmpe, filter_imt=filter_imt
                )
                if out is None:
                    return None
            else:
                raise TypeError(
                    f"{selected_gmpe} must consist exclusively of keys in "
                    "conf['gmpe_modules'] or conf['gmpe_sets']"
                )
        elif selected_gmpe in conf["gmpe_modules"].keys():
            modinfo = conf["gmpe_modules"][selected_gmpe]
            # mod = import_module(modinfo[1])
            # tmpclass = getattr(mod, modinfo[0])
            # out = MultiGMPE.__from_list__([tmpclass()], [1.0], imc=imc)
            out = MultiGMPE.__from_list__(
                [get_gmpe_from_name(modinfo[0], conf)],
                [1.0],
                imc=imc,
                imt=filter_imt,
            )
            if out is None:
                return None
        else:
            raise TypeError(
                "conf['modeling']['gmpe'] must be a key in "
                "conf['gmpe_modules'] or conf['gmpe_sets']"
            )

        out.DESCRIPTION = selected_gmpe

        # ---------------------------------------------------------------------
        # Deal with GMPE limits
        # ---------------------------------------------------------------------
        gmpe_lims = conf["gmpe_limits"]

        # We need to replace the short name in the dictionary key with module
        # name here since the conf is not available within the MultiGMPE class.
        mods = conf["gmpe_modules"]
        mod_keys = mods.keys()
        new_gmpe_lims = {}
        for k, v in gmpe_lims.items():
            if k in mod_keys:
                new_gmpe_lims[mods[k][0]] = v
            else:
                new_gmpe_lims[k] = v

        out.GMPE_LIMITS = new_gmpe_lims

        return out

    @classmethod
    def __multigmpe_from_gmpe_set__(cls, conf, set_name, filter_imt=None):
        """
        Private method for constructing a MultiGMPE from a set_name.

        Args:
            conf (dict): A ShakeMap ConfigObj config object.
            filter_imt (IMT): An optional IMT to filter/reweight the GMPE list.
            set_name (str): Set name; must correspond to a key in
                conf['set_name'].

        Returns:
            MultiGMPE.

        """
        imc = getattr(const.IMC, conf["interp"]["component"])

        selected_gmpes = conf["gmpe_sets"][set_name]["gmpes"]
        selected_gmpe_weights = [
            float(w) for w in conf["gmpe_sets"][set_name]["weights"]
        ]

        # ---------------------------------------------------------------------
        # Import GMPE modules and initialize classes into list
        # ---------------------------------------------------------------------
        gmpes = []
        for g in selected_gmpes:
            # This is the old school way of importing the modules; I'm
            # leaving it in here temporarily just for documentation.
            # mod = import_module(conf['gmpe_modules'][g][1])
            # tmpclass = getattr(mod, conf['gmpe_modules'][g][0])
            # gmpes.append(tmpclass())
            gmpe_name = conf["gmpe_modules"][g][0]
            gmpes.append(get_gmpe_from_name(gmpe_name, conf))

        # ---------------------------------------------------------------------
        # Filter out GMPEs not applicable to this period
        # ---------------------------------------------------------------------
        if filter_imt is not None:
            filtered_gmpes, filtered_wts = filter_gmpe_list(
                gmpes, selected_gmpe_weights, filter_imt
            )
        else:
            filtered_gmpes, filtered_wts = gmpes, selected_gmpe_weights

        # ---------------------------------------------------------------------
        # Construct MultiGMPE
        # ---------------------------------------------------------------------
        logging.debug("    filtered_gmpes: %s", filtered_gmpes)
        logging.debug("    filtered_wts: %s", filtered_wts)

        if len(filtered_gmpes) == 0:
            return None

        mgmpe = MultiGMPE.__from_list__(
            filtered_gmpes,
            filtered_wts,
            imc=imc,
        )

        mgmpe.DESCRIPTION = set_name

        return mgmpe

    @classmethod
    def __from_list__(
        cls,
        gmpes,
        weights,
        imc=const.IMC.GREATER_OF_TWO_HORIZONTAL,
        imt=None,
        reference_vs30=760,
    ):
        """
        Construct a MultiGMPE instance from lists of GMPEs and weights.

        Args:
            gmpes (list): List of OpenQuake
                `GMPE <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#built-in-gsims>`__
                instances.

            weights (list): List of weights; must sum to 1.0.

            imc: Requested intensity measure component. Must be one listed
                `here <http://docs.openquake.org/oq-hazardlib/master/const.html?highlight=imc#openquake.hazardlib.const.IMC>`__.
                The amplitudes returned by the GMPEs will be converted to this
                IMC. Default is 'GREATER_OF_TWO_HORIZONTAL', which is used by
                ShakeMap. See discussion in
                `this section <http://usgs.github.io/shakemap/tg_choice_of_parameters.html#use-of-peak-values-rather-than-mean>`__
                of the ShakeMap manual.

            reference_vs30:
                Reference rock Vs30 in m/s. We do not check that this matches
                the reference rock in the GMPEs so this is the responsibility
                of the user.

        """  # noqa

        # ---------------------------------------------------------------------
        # Check that GMPE weights sum to 1.0:
        # ---------------------------------------------------------------------

        if np.abs(np.sum(weights) - 1.0) > 1e-7:
            raise ValueError("Weights must sum to one.")

        # ---------------------------------------------------------------------
        # Check that length of GMPE weights equals length of gmpe list
        # ---------------------------------------------------------------------

        if len(weights) != len(gmpes):
            raise ValueError(
                "Length of weights must match length of GMPE list."
            )

        # ---------------------------------------------------------------------
        # Check that gmpes is a list of OQ GMPE instances
        # ---------------------------------------------------------------------

        for g in gmpes:
            if not isinstance(g, GMPE):
                raise ValueError(f'"{g}" is a {type(g)} not a GMPE instance.')

        if imt is not None:
            gmpes, weights = filter_gmpe_list(gmpes, weights, imt)
            if len(gmpes) == 0:
                return None

        self = cls()
        self.GMPES = gmpes
        self.WEIGHTS = weights

        # ---------------------------------------------------------------------
        # Combine the intensity measure types. This is problematic:
        #   - Logically, we should only include the intersection of the sets
        #     of imts for the different GMPEs.
        #   - In practice, this is not feasible because most GMPEs in CEUS and
        #     subduction zones do not have PGV.
        #   - So instead we will use the union of the imts and then convert
        #     to get the missing imts later in get_mean_and_stddevs.
        # ---------------------------------------------------------------------

        imts = [set(g.DEFINED_FOR_INTENSITY_MEASURE_TYPES) for g in gmpes]
        self.DEFINED_FOR_INTENSITY_MEASURE_TYPES = set.union(*imts)

        # ---------------------------------------------------------------------
        # For VirtualIPE class, we also want to know if ALL of the GMPEs are
        # defined for PGV, in which case we will convert from PGV to MI,
        # otherwise use PGA or Sa.
        # ---------------------------------------------------------------------
        haspgv = [
            any(
                "PGV" in imtype.__name__
                for imtype in g.DEFINED_FOR_INTENSITY_MEASURE_TYPES
            )
            for g in gmpes
        ]
        self.ALL_GMPES_HAVE_PGV = all(haspgv)

        # ---------------------------------------------------------------------
        # Store intensity measure types for conversion in get_mean_and_stddevs.
        # ---------------------------------------------------------------------
        self.IMCs = [g.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT for g in gmpes]

        # ---------------------------------------------------------------------
        # Store the component
        # ---------------------------------------------------------------------
        self.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = imc

        # ---------------------------------------------------------------------
        # Intersection of GMPE standard deviation types
        # ---------------------------------------------------------------------
        stdlist = [set(g.DEFINED_FOR_STANDARD_DEVIATION_TYPES) for g in gmpes]
        self.DEFINED_FOR_STANDARD_DEVIATION_TYPES = set.intersection(*stdlist)

        # ---------------------------------------------------------------------
        # Need union of site parameters, but it is complicated by the
        # different depth parameter flavors.
        # ---------------------------------------------------------------------
        sitepars = [set(g.REQUIRES_SITES_PARAMETERS) for g in gmpes]
        self.REQUIRES_SITES_PARAMETERS = set.union(*sitepars)

        # ---------------------------------------------------------------------
        # Construct a list of whether or not each GMPE has a site term
        # ---------------------------------------------------------------------
        self.HAS_SITE = ["vs30" in g.REQUIRES_SITES_PARAMETERS for g in gmpes]

        self.REFERENCE_VS30 = reference_vs30

        # ---------------------------------------------------------------------
        # Union of rupture parameters
        # ---------------------------------------------------------------------
        ruppars = [set(g.REQUIRES_RUPTURE_PARAMETERS) for g in gmpes]
        self.REQUIRES_RUPTURE_PARAMETERS = set.union(*ruppars)

        # ---------------------------------------------------------------------
        # Union of distance parameters
        # ---------------------------------------------------------------------
        distpars = [set(g.REQUIRES_DISTANCES) for g in gmpes]
        self.REQUIRES_DISTANCES = set.union(*distpars)

        return self

    def __describe__(self):
        """
        Construct a dictionary that describes the MultiGMPE.

        Note: For simplicity, this method ignores issues related to
        GMPEs used for the site term and changes in the GMPE with
        distance. For this level of detail, please see the config files.

        Returns:
            A dictionary representation of the MultiGMPE.
        """
        gmpe_dict = {"gmpes": [], "weights": [], "name": self.DESCRIPTION}

        for i, this_gmpe in enumerate(self.GMPES):
            gmpe_dict["weights"].append(self.WEIGHTS[i])
            if isinstance(this_gmpe, MultiGMPE):
                gmpe_dict["gmpes"].append(this_gmpe.__describe__())
            else:
                gmpe_dict["gmpes"].append(str(this_gmpe))

        return gmpe_dict


def filter_gmpe_list(gmpes, wts, imt):
    """
    Method to remove GMPEs from the GMPE list that are not applicable
    to a specific IMT. Rescales the weights to sum to one.

    Args:
        gmpes (list): List of GMPE instances.
        wts (list): List of floats indicating the weight of the GMPEs.
        imt (IMT): OQ IMT to filter GMPE list for.

    Returns:
        tuple: List of GMPE instances and list of weights.

    """
    if wts is None:
        n = len(gmpes)
        wts = [1 / n] * n

    per_max = [np.max(get_gmpe_sa_periods(g)) for g in gmpes]
    per_min = [np.min(get_gmpe_sa_periods(g)) for g in gmpes]
    if imt == PGA():
        sgmpe = [
            g
            for g in gmpes
            if any(
                "PGA" in t.__name__
                for t in g.DEFINED_FOR_INTENSITY_MEASURE_TYPES
            )
        ]
        swts = [
            w
            for g, w in zip(gmpes, wts)
            if any(
                "PGA" in t.__name__
                for t in g.DEFINED_FOR_INTENSITY_MEASURE_TYPES
            )
        ]
    elif imt == PGV():
        sgmpe = []
        swts = []
        for i, this_gmpe in enumerate(gmpes):
            if (
                any(
                    "PGV" in t.__name__
                    for t in this_gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES
                )
            ) or (per_max[i] >= 1.0 >= per_min[i]):
                sgmpe.append(this_gmpe)
                swts.append(wts[i])
    else:
        per = imt.period
        sgmpe = []
        swts = []
        for i, this_gmpe in enumerate(gmpes):
            if per_max[i] >= per >= per_min[i]:
                sgmpe.append(this_gmpe)
                swts.append(wts[i])

    if len(sgmpe) == 0:
        return [], []
    #    raise KeyError(f"No applicable GMPEs from GMPE list for {str(imt)}")

    # Scale weights to sum to one
    swts = np.array(swts)
    swts = swts / np.sum(swts)

    return sgmpe, swts



#!/usr/bin/env python

# stdlib imports
import copy

import esi_utils_vectors.ecef as ecef

# third party imports
import numpy as np
import openquake.hazardlib.geo as geo
from esi_utils_rupture.distance import get_distance
from esi_utils_rupture.utils import _distance_sq_to_segment
from esi_utils_vectors.vector import Vector


class Bayless2013(object):
    """
    Implements the Bayless and Somerville (2013) directivity model. This is an
    update to Somerville et al. (1997), which defines some of the input
    parameters.

    Fd is the directivity effect parameter, and it is evaluated as

    Fd = (c0 + c1 * Fgeom) * Tcd * Tmw * Taz

    Note that Fd is intended to be used in GMPEs as:

    ln(IM_dir) = ln(IM) + Fd

    where

    - Fd is the directivity effect
    - IM is the intensity measure predicted by the GMPE.
    - IM_dir is the directivity-adjusted IM.

    The model is seprated into strike-slip and dip-slip categories:
    SS: (abs(rake) >  0 & abs(rake) < 30) | (abs(rake) > 150 & abs(rake) < 180)
    DS:  abs(rake) > 60 & abs(rake) < 120

    Notes from Somerville et al. (1997):

    * d = length of dipping fault rupturing towards site
    * Y = d/W
    * s = length of striking fault rupturing towards site
    * X = s/L

    Bayless and Somerville state that each quadrilateral should have a pseudo-
    hypocenter:

        "Define the pseudo-hypocenter for rupture of successive segments as the
        point on the side edge of the fault segment that is closest to the side
        edge of the previous segment, and that lies half way between the top
        and bottom of the fault. We assume that the fault is segmented along
        strike, not updip. All geometric parameters are computed relative to
        the pseudo-hypocenter."

    To do:

        - Interpolate for arbitrary periods
        - Inherit from Directivity base class.

    References:
        Bayless, J., and Somerville, P. (2013). Bayless-Somerville Directivity
        Model, Chapter 3 of PEER Report No. 2013/09, P. Spudich (Editor),
        Pacific Earthquake Engineering Research Center, Berkeley, CA.
        `[link] <http://peer.berkeley.edu/publications/peer_reports/reports_2013/webPEER-2013-09-Spudich.pdf>`__

        Somerville, P. G., Smith, N. F., Graves, R. W., and Abrahamson, N. A.
        (1997). Modification of empirical strong ground motion attenuation
        relations to include the amplitude and duration effects of rupture
        directivity, Seismol. Res. Lett. 68, 199–222.
        `[link] <http://srl.geoscienceworld.org/content/68/1/199.abstract>`__

    """  # noqa

    # ------------------------------------------------------------------------
    # C0 adn C1 are for RotD50. One set for each mechanism (SS vs DS).
    # FN and FP are also available
    # ------------------------------------------------------------------------
    __periods = np.array([0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 4.0, 5.0, 7.5, 10])
    __c0ss = np.array(
        [0.0, 0.0, -0.12, -0.175, -0.21, -0.235, -0.255, -0.275, -0.29, -0.3]
    )
    __c1ss = np.array(
        [0.0, 0.0, 0.075, 0.090, 0.095, 0.099, 0.103, 0.108, 0.112, 0.115]
    )
    __c0ds = np.array([0.0, 0.0, 0.0, 0.0, 0.0, -0.033, -0.089, -0.133, -0.16, -0.176])
    __c1ds = np.array([0.0, 0.0, 0.0, 0.0, 0.034, 0.093, 0.128, 0.15, 0.165, 0.179])

    def __init__(self, origin, rup, lat, lon, depth, T):
        """
        Constructor for bayless2013.

        Args:
            origin: Shakemap Origin object.
            rup: Shakemap Rupture object.
            lat: Numpy array of latitudes.
            lon: Numpy array of longitudes.
            depth: Numpy array of depths (km); down is positive.
            T: Period; Currently, only acceptable values are
                0.5, 0.75, 1, 1.5, 2, 3, 4, 5, 7.5, 10.
        """
        self._origin = origin
        self._rup = rup
        self._rake = origin.rake
        self._M = origin.mag
        self._hyp = origin.getHypo()
        self._lon = lon
        self._lat = lat
        self._dep = depth
        self._T = T

        # Lists of widths and lengths for each quad in the rupture
        self._W = self._rup.getIndividualWidths()
        self._L = self._rup.getIndividualTopLengths()

        # Number of quads
        self._nq = len(self._W)

        # Currently assuming that the rake is the same on all subruptures .
        self.__getSlipCategory()

        # Rupture weights are supposed to be based on seismic moment.
        # Since moment is proportional to area, lets just use area
        # for now
        area = self._W * self._L
        self.weights = area / np.sum(area)

        # Put in pseudo-hypocenters for each quad
        self.__setPseudoHypocenters()

        self._fd = 0
        for i in range(self._nq):
            self.i = i

            # Compute some genral stuff that is required for all mechanisms
            dtypes = ["rrup", "rx", "ry0"]
            dists = get_distance(dtypes, self._lat, self._lon, self._dep, self._rup)
            self.__Rrup = np.reshape(dists["rrup"], self._lat.shape)
            self.__Rx = np.reshape(dists["rx"], self._lat.shape)
            self.__Ry = np.reshape(dists["ry0"], self._lat.shape)
            # NOTE: use Rx and Ry to compute Az in 'computeAz'. It is probably
            #       possible to make this a lot faster by avoiding the
            #       calculation of these distances each time.

            # Az is the NGA definition of source-to-site azimuth for a finite
            # rupture. See Kaklamanos et al. (2011) Figure 2 for illustration.

            self.__computeAz()  # uses Rx and Ry, which are for the i-th quad.

            # Magnitude taper (does not depend on mechanism)
            if self._M <= 5.0:
                self._T_Mw = 0.0
            elif (self._M > 5.0) and (self._M < 6.5):
                self._T_Mw = 1.0 - (6.5 - self._M) / 1.5
            else:
                self._T_Mw = 1.0

            if self.SlipCategory == "SS":
                self.__computeSS()
                self._fd = self._fd + self.weights[i] * self._fd_SS

            elif self.SlipCategory == "DS":
                self.__computeDS()
                self._fd = self._fd + self.weights[i] * self._fd_DS

            else:
                # Compute both SS and DS
                self.__computeSS()
                self.__computeDS()

                # Normalize rake to reference angle
                sintheta = np.abs(np.sin(np.radians(self._rake)))
                costheta = np.abs(np.cos(np.radians(self._rake)))
                refrake = np.arctan2(sintheta, costheta)

                # Compute weights:
                DipWeight = refrake / (np.pi / 2.0)
                StrikeWeight = 1.0 - DipWeight
                fdcombined = StrikeWeight * self._fd_SS + DipWeight * self._fd_DS
                self._fd = self._fd + self.weights[i] * fdcombined

    def __setPseudoHypocenters(self):
        """Set a pseudo-hypocenter.

        Adapted from ShakeMap 3.5 src/contour/directivity.c

        From Bayless and Somerville:

        "Define the pseudo-hypocenter for rupture of successive segments as
        the point on the side edge of the rupture segment that is closest to
        the side edge of the previous segment, and that lies half way
        between the top and bottom of the rupture. We assume that the rupture
        is segmented along strike, not updip. All geometric parameters are
        computed relative to the pseudo-hypocenter."
        """
        hyp_ecef = Vector.fromPoint(
            geo.point.Point(self._hyp.longitude, self._hyp.latitude, self._hyp.depth)
        )
        # Loop over each quad
        self.phyp = [None] * self._nq
        for i in range(self._nq):
            P0, P1, P2, P3 = self._rup.getQuadrilaterals()[i]
            p0 = Vector.fromPoint(P0)  # convert to ECEF
            p1 = Vector.fromPoint(P1)
            p2 = Vector.fromPoint(P2)
            p3 = Vector.fromPoint(P3)

            # Create 4 planes with normals pointing outside rectangle
            hpnp = Vector.cross(p1 - p0, p2 - p0).norm()
            hpp = -hpnp.x * p0.x - hpnp.y * p0.y - hpnp.z * p0.z
            n0 = Vector.cross(p1 - p0, hpnp)
            n1 = Vector.cross(p2 - p1, hpnp)
            n2 = Vector.cross(p3 - p2, hpnp)
            n3 = Vector.cross(p0 - p3, hpnp)

            # -----------------------------------------------------------------
            # Is the hypocenter inside the projected rectangle?
            # Dot products show which side the origin is on.
            # If origin is on same side of all the planes, then it is 'inside'
            # -----------------------------------------------------------------

            sgn0 = np.signbit(Vector.dot(n0, p0 - hyp_ecef))
            sgn1 = np.signbit(Vector.dot(n1, p1 - hyp_ecef))
            sgn2 = np.signbit(Vector.dot(n2, p2 - hyp_ecef))
            sgn3 = np.signbit(Vector.dot(n3, p3 - hyp_ecef))

            if (sgn0 == sgn1) and (sgn1 == sgn2) and (sgn2 == sgn3):
                # Origin is inside. Use distance-to-plane formula.
                d = Vector.dot(hpnp, hyp_ecef) + hpp
                d = d * d

                # Put the pseudo hypocenter on the plane
                D = Vector.dot(hpnp, hyp_ecef) + hpp
                self.phyp[i] = hyp_ecef - hpnp * D

            else:
                # Origin is outside. Find distance to edges
                p0p = np.reshape(p0.getArray() - hyp_ecef.getArray(), [1, 3])
                p1p = np.reshape(p1.getArray() - hyp_ecef.getArray(), [1, 3])
                p2p = np.reshape(p2.getArray() - hyp_ecef.getArray(), [1, 3])
                p3p = np.reshape(p3.getArray() - hyp_ecef.getArray(), [1, 3])
                s1 = _distance_sq_to_segment(p1p, p2p)
                s3 = _distance_sq_to_segment(p3p, p0p)

                # Assuming that the rupture is segmented along strike and not
                # updip (as described by Bayless and somerville), we only
                # need to consider s1 and s3:
                if s1 > s3:
                    e30 = p0 - p3
                    e30norm = e30.norm()
                    mag = e30.mag()
                    self.phyp[i] = p3 + e30norm * (0.5 * mag)
                else:
                    e21 = p1 - p2
                    e21norm = e21.norm()
                    mag = e21.mag()
                    self.phyp[i] = p2 + e21norm * (0.5 * mag)

    def __computeDS(self):
        # d is the length of dipping rupture rupturing toward site;
        # Note: max[(Y*W),exp(0)] -- just apply a min of 1?
        self.__computeD(self.i)

        # Geometric directivity predictor:
        RxoverW = (self.__Rx / self._W[self.i]).clip(
            min=-np.pi / 2.0, max=2.0 * np.pi / 3.0
        )
        f_geom = np.log(self.d) * np.cos(RxoverW)

        # Distance taper
        T_CD = np.ones_like(self._lat)
        ix = tuple(
            [
                (self.__Rrup / self._W[self.i] > 1.5)
                & (self.__Rrup / self._W[self.i] < 2.0)
            ]
        )
        T_CD[ix] = 1.0 - (self.__Rrup[ix] / self._W[self.i] - 1.5) / 0.5
        T_CD[self.__Rrup / self._W[self.i] >= 2.0] = 0.0

        # Azimuth taper
        T_Az = np.sin(np.abs(self.Az)) ** 2

        # Select Coefficients
        ix = tuple([self._T == self.__periods])
        C0 = self.__c0ds[ix]
        C1 = self.__c1ds[ix]

        self._fd_DS = (C0 + C1 * f_geom) * T_CD * self._T_Mw * T_Az

    def __computeSS(self):
        # s is the length of striking fault rupturing toward site;
        # max[(X*L),exp(1)]
        # theta (see Figure 5 in SSGA97)
        self.__computeThetaAndS(self.i)

        # Geometric directivity predictor:
        f_geom = np.log(self.s) * (0.5 * np.cos(2 * self.theta) + 0.5)

        # Distance taper
        T_CD = np.ones_like(self._lat)
        ix = tuple(
            [
                (self.__Rrup / self._L[self.i] > 0.5)
                & (self.__Rrup / self._L[self.i] < 1.0)
            ]
        )
        T_CD[ix] = 1 - (self.__Rrup[ix] / self._L[self.i] - 0.5) / 0.5
        T_CD[self.__Rrup / self._L[self.i] >= 1.0] = 0.0

        # Azimuth taper
        T_Az = 1.0

        # Select Coefficients
        ix = tuple([self._T == self.__periods])
        C0 = self.__c0ss[ix]
        C1 = self.__c1ss[ix]
        self._fd_SS = (C0 + C1 * f_geom) * T_CD * self._T_Mw * T_Az

    def __computeAz(self):
        Az = np.ones_like(self.__Rx) * np.pi / 2.0
        Az = Az * np.sign(self.__Rx)
        ix = tuple([self.__Ry > 0.0])
        Az[ix] = np.arctan(self.__Rx[ix] / self.__Ry[ix])
        self.Az = Az

    def __computeD(self, i):
        """Compute d for the i-th quad/segment.

        Y = d/W, where d is the portion (in km) of the width of the fault which
        ruptures up-dip from the hypocenter to the top of the fault.

        Args:
            i (int): index of segment for which d is to be computed.
        """
        hyp_ecef = self.phyp[i]  # already in ECEF
        hyp_col = np.array([[hyp_ecef.x], [hyp_ecef.y], [hyp_ecef.z]])

        # First compute "updip" vector
        P0, P1, P2 = self._rup.getQuadrilaterals()[i][0:3]
        p1 = Vector.fromPoint(P1)  # convert to ECEF
        p2 = Vector.fromPoint(P2)
        e21 = p1 - p2
        e21norm = e21.norm()
        hp1 = p1 - hyp_ecef
        # convert to km (used as max later)
        udip_len = Vector.dot(hp1, e21norm) / 1000.0
        udip_col = np.array([[e21norm.x], [e21norm.y], [e21norm.z]])  # ECEF coords

        # Sites
        slat = self._lat
        slon = self._lon

        # Convert sites to ECEF:
        site_ecef_x = np.ones_like(slat)
        site_ecef_y = np.ones_like(slat)
        site_ecef_z = np.ones_like(slat)

        # Make a 3x(#number of sites) matrix of site locations
        # (rows are x, y, z) in ECEF
        site_ecef_x, site_ecef_y, site_ecef_z = ecef.latlon2ecef(
            slat, slon, np.zeros(slon.shape)
        )
        site_mat = np.array(
            [
                np.reshape(site_ecef_x, (-1,)),
                np.reshape(site_ecef_y, (-1,)),
                np.reshape(site_ecef_z, (-1,)),
            ]
        )

        # Hypocenter-to-site matrix
        h2s_mat = site_mat - hyp_col  # in ECEF

        # Dot hypocenter-to-site with updip vector
        d_raw = np.abs(np.sum(h2s_mat * udip_col, axis=0)) / 1000.0  # convert to km
        d_raw = np.reshape(d_raw, self._lat.shape)
        self.d = d_raw.clip(min=1.0, max=udip_len)

    def __computeThetaAndS(self, i):
        """
        Args:
            i (int): Compute d for the i-th quad/segment.
        """
        # self.phyp is in ECEF
        tmp = ecef.ecef2latlon(self.phyp[i].x, self.phyp[i].y, self.phyp[i].z)
        epi_ecef = Vector.fromPoint(geo.point.Point(tmp[1], tmp[0], 0.0))
        epi_col = np.array([[epi_ecef.x], [epi_ecef.y], [epi_ecef.z]])

        # First compute along strike vector
        P0, P1, P2, P3 = self._rup.getQuadrilaterals()[i]
        p0 = Vector.fromPoint(P0)  # convert to ECEF
        p1 = Vector.fromPoint(P1)
        e01 = p1 - p0
        e01norm = e01.norm()
        hp0 = p0 - epi_ecef
        hp1 = p1 - epi_ecef
        strike_min = Vector.dot(hp0, e01norm) / 1000.0  # convert to km
        strike_max = Vector.dot(hp1, e01norm) / 1000.0  # convert to km
        strike_col = np.array([[e01norm.x], [e01norm.y], [e01norm.z]])  # ECEF coords

        # Sites
        slat = self._lat
        slon = self._lon

        # Convert sites to ECEF:
        site_ecef_x = np.ones_like(slat)
        site_ecef_y = np.ones_like(slat)
        site_ecef_z = np.ones_like(slat)

        # Make a 3x(#number of sites) matrix of site locations
        # (rows are x, y, z) in ECEF
        site_ecef_x, site_ecef_y, site_ecef_z = ecef.latlon2ecef(
            slat, slon, np.zeros(slon.shape)
        )
        site_mat = np.array(
            [
                np.reshape(site_ecef_x, (-1,)),
                np.reshape(site_ecef_y, (-1,)),
                np.reshape(site_ecef_z, (-1,)),
            ]
        )

        # Epicenter-to-site matrix
        e2s_mat = site_mat - epi_col  # in ECEF
        mag = np.sqrt(np.sum(e2s_mat * e2s_mat, axis=0))

        # Avoid division by zero
        mag[mag == 0] = 1e-12
        e2s_norm = e2s_mat / mag

        # Dot epicenter-to-site with along-strike vector
        s_raw = np.sum(e2s_mat * strike_col, axis=0) / 1000.0  # conver to km

        # Put back into a 2d array
        s_raw = np.reshape(s_raw, self._lat.shape)
        self.s = np.abs(s_raw.clip(min=strike_min, max=strike_max)).clip(min=np.exp(1))

        # Compute theta
        sdots = np.sum(e2s_norm * strike_col, axis=0)
        theta_raw = np.arccos(sdots)

        # But theta is defined to be the reference angle
        # (i.e., the equivalent angle between 0 and 90 deg)
        sintheta = np.abs(np.sin(theta_raw))
        costheta = np.abs(np.cos(theta_raw))
        theta = np.arctan2(sintheta, costheta)
        self.theta = np.reshape(theta, self._lat.shape)

    def getFd(self):
        """
        Returns:
            Numpy array of Fd; Fd is the directivity amplification factor.
        """
        return copy.deepcopy(self._fd)

    def __getSlipCategory(self):
        """
        Sets self.SlipCategory based on rake angle. Can be SS for
        strike slip, DS for dip slip, or Unspecified.
        """
        arake = np.abs(self._rake)
        self.SlipCategory = "Unspecified"
        if ((arake >= 0) and (arake <= 30)) or ((arake >= 150) and (arake <= 180)):
            self.SlipCategory = "SS"
        if (arake >= 60) and (arake <= 120):
            self.SlipCategory = "DS"



#!/usr/bin/env python

# stdlib imports

# third party imports
from mapio.reader import read, get_file_geodict
from mapio.grid2d import Grid2D
from mapio.geodict import GeoDict

from esi_shakelib.utils.contexts import SitesContext
import numpy as np

# local imports
from esi_shakelib.utils.exception import ShakeLibException


class Sites(object):
    """
    An object to encapsulate information used to generate a GEM
    `SitesContext <https://github.com/gem/oq-hazardlib/blob/master/openquake/hazardlib/gsim/base.py>`__.
    """  # noqa

    def __init__(
        self, vs30grid, vs30measured_grid=None, backarc=None, defaultVs30=760.0
    ):
        """
        Construct a Sites object.

        Args:
            vs30grid: MapIO Grid2D object containing Vs30 values.
            vs30measured_grid: Boolean array indicating whether Vs30 values
                were measured or derived (i.e., from topographic slope).
            backarc: Boolean array indicating whether site is in the subduction
                `backarc <http://earthquake.usgs.gov/learn/glossary/?term=backarc>`__.
            defaultVs30: Default Vs30 value to use in locations where Vs30Grid
                is not specified.
        """  # noqa
        self._Vs30 = vs30grid
        if backarc is None:
            self._backarc = np.zeros_like(vs30grid.getData(), dtype=bool)
        else:
            self._backarc = backarc
            # Could add a check here that if backarc is provided that it's type
            # is bool and dimensions match vs30grid

        self._defaultVs30 = defaultVs30
        self._vs30measured_grid = vs30measured_grid
        self._GeoDict = vs30grid.getGeoDict().copy()
        if self._GeoDict.xmin > self._GeoDict.xmax:
            xmin = self._GeoDict.xmin - 360
            xmax = self._GeoDict.xmax
        else:
            xmin = self._GeoDict.xmin
            xmax = self._GeoDict.xmax
        self._lons = np.linspace(xmin, xmax, self._GeoDict.nx)
        self._lats = np.linspace(
            self._GeoDict.ymax, self._GeoDict.ymin, self._GeoDict.ny
        )
        # _lons = np.linspace(xmin, xmax, self._GeoDict.nx)
        # _lats = np.linspace(self._GeoDict.ymax, self._GeoDict.ymin, self._GeoDict.ny)
        # lons, lats = np.meshgrid(_lons, _lats)
        # self._lons = lons
        # self._lats = lats

    @classmethod
    def _create(cls, geodict, defaultVs30, vs30File, padding, resample):
        if vs30File is not None:
            fgeodict = get_file_geodict(vs30File)
            if not resample:
                if not padding:
                    # we want something that is within and aligned
                    geodict = fgeodict.getBoundsWithin(geodict)
                else:
                    # we want something that is just aligned, since we're
                    # padding edges
                    geodict = fgeodict.getAligned(geodict)
            vs30grid = read(
                vs30File,
                samplegeodict=geodict,
                resample=resample,
                method="linear",
                doPadding=padding,
                padValue=defaultVs30,
            )
        return vs30grid

    @classmethod
    def fromBounds(
        cls,
        xmin,
        xmax,
        ymin,
        ymax,
        dx,
        dy,
        defaultVs30=760.0,
        vs30File=None,
        vs30measured_grid=None,
        backarc=None,
        padding=False,
        resample=False,
    ):
        """
        Create a Sites object by defining a center point, resolution, extent,
        and Vs30 values.

        Args:
            xmin: X coordinate of left edge of bounds.
            xmax: X coordinate of right edge of bounds.
            ymin: Y coordinate of bottom edge of bounds.
            ymax: Y coordinate of top edge of bounds.
            dx: Resolution of desired grid in X direction.
            dy: Resolution of desired grid in Y direction.
            defaultVs30: Default Vs30 value to use if vs30File not specified.
            vs30File: Name of GMT or GDAL format grid file containing Vs30
                values.
            vs30measured_grid: Boolean grid indicating whether Vs30 values were
                measured or derived (i.e., from slope).
            backarc: Boolean array indicating whether site is in the subduction
                `backarc <http://earthquake.usgs.gov/learn/glossary/?term=backarc>`__.
            padding: Boolean indicating whether or not to pad resulting Vs30
                grid out to edges of input bounds. If False, grid will be
                clipped to the extent of the input file.
            resample: Boolean indicating whether or not the grid should be
                resampled.
        """  # noqa
        geodict = GeoDict.createDictFromBox(xmin, xmax, ymin, ymax, dx, dy)
        if vs30File is not None:
            vs30grid = cls._create(geodict, defaultVs30, vs30File, padding, resample)
        else:
            griddata = np.ones((geodict.ny, geodict.nx), dtype=np.float64) * defaultVs30
            vs30grid = Grid2D(griddata, geodict)
        return cls(
            vs30grid,
            vs30measured_grid=vs30measured_grid,
            backarc=backarc,
            defaultVs30=defaultVs30,
        )

    @classmethod
    def fromCenter(
        cls,
        cx,
        cy,
        xspan,
        yspan,
        dx,
        dy,
        defaultVs30=760.0,
        vs30File=None,
        vs30measured_grid=None,
        backarc=None,
        padding=False,
        resample=False,
    ):
        """
        Create a Sites object by defining a center point, resolution, extent,
        and Vs30 values.

        Args:
            cx: X coordinate of desired center point.
            cy: Y coordinate of desired center point.
            xspan: Width of desired grid.
            yspan: Height of desired grid.
            dx: Resolution of desired grid in X direction.
            dy: Resolution of desired grid in Y direction.
            defaultVs30: Default Vs30 value to use if vs30File not specified.
            vs30File: Name of GMT or GDAL format grid file containing Vs30
                values.
            vs30measured_grid: Boolean grid indicating whether Vs30 values were
                measured or derived (i.e., from slope).
            backarc: Boolean array indicating whether site is in the subduction
                `backarc <http://earthquake.usgs.gov/learn/glossary/?term=backarc>`__.
            padding: Boolean indicating whether or not to pad resulting Vs30
                grid out to edges of input bounds. If False, grid will be
                clipped to the extent of the input file.
            resample: Boolean indicating whether or not the grid should be
                resampled.
        """  # noqa
        geodict = GeoDict.createDictFromCenter(cx, cy, dx, dy, xspan, yspan)
        if vs30File is not None:
            vs30grid = cls._create(geodict, defaultVs30, vs30File, padding, resample)
        else:
            griddata = np.ones((geodict.ny, geodict.nx), dtype=np.float64) * defaultVs30
            vs30grid = Grid2D(griddata, geodict)
        return cls(
            vs30grid,
            vs30measured_grid=vs30measured_grid,
            backarc=backarc,
            defaultVs30=defaultVs30,
        )

    def getSitesContext(self, lldict=None, rock_vs30=None):
        """
        Create a SitesContext object by sampling the current Sites object.

        Args:
            lldict: Either

                - None, in which case the SitesContext for the complete Sites
                  grid is returned, or
                - A location dictionary (elements are 'lats' and 'lons' and
                  each is a numpy array). Each element must have the same
                  shape. In this case the SitesContext for these locaitons is
                  returned.

            rock_vs30: Either

                - None, in which case the SitesContext will reflect the Vs30
                  grid in the Sites instance, or
                - A float for the rock Vs30 value, in which case the
                  SitesContext will be constructed for this constant Vs30
                  value.

        Returns:
            SitesContext object.

        Raises:
            ShakeLibException: When lat/lon input sequences do not share
                dimensionality.

        """  # noqa
        sctx = SitesContext()

        if lldict is not None:
            lats = lldict["lats"]
            lons = lldict["lons"]
            latshape = lats.shape
            lonshape = lons.shape
            if latshape != lonshape:
                msg = "Input lat/lon arrays must have the same dimensions"
                raise ShakeLibException(msg)

            if rock_vs30 is not None:
                tmp = self._Vs30.getValue(lats, lons, default=self._defaultVs30)
                sctx.vs30 = np.ones_like(tmp) * rock_vs30
            else:
                sctx.vs30 = self._Vs30.getValue(lats, lons, default=self._defaultVs30)
            sctx.lats = lats.copy()
            sctx.lons = lons.copy()
            sctx.sids = np.array(range(np.size(lons))).reshape(lons.shape)
        else:
            sctx.lons = self._lons.copy()
            sctx.lats = self._lats.copy()
            if rock_vs30 is not None:
                sctx.vs30 = np.full_like(self._Vs30.getData(), rock_vs30)
            else:
                sctx.vs30 = self._Vs30.getData().copy()
            sctx.sids = np.array(
                range(np.size(self._lons) * np.size(self._lats))
            ).reshape(sctx.vs30.shape)

        addDepthParameters(sctx)

        # For ShakeMap purposes, vs30 measured is always Fales
        sctx.vs30measured = np.zeros_like(sctx.vs30, dtype=bool)

        # Backarc should be a numpy array
        if lldict is not None:
            tgd = self._Vs30.getGeoDict()
            backarcgrid = Grid2D(self._backarc, tgd)
            sctx.backarc = backarcgrid.getValue(lats, lons, default=False)
        else:
            sctx.backarc = self._backarc.copy()

        return sctx

    def getVs30Grid(self):
        """

        Returns: Grid2D object containing Vs30 values for this Sites object.
        """
        return self._Vs30

    def getNxNy(self):
        """
        Returns: The number of grid points in x and y.
        """
        return self._GeoDict.nx, self._GeoDict.ny

    @staticmethod
    def _load(
        vs30File,
        samplegeodict=None,
        resample=False,
        method="linear",
        doPadding=False,
        padValue=np.nan,
    ):
        try:
            vs30grid = read(
                vs30File,
                samplegeodict=samplegeodict,
                resample=resample,
                method=method,
                doPadding=doPadding,
                padValue=padValue,
            )
        except Exception as msg1:
            msg = f'Load failure of {vs30File} - error message: "{str(msg1)}"'
            raise ShakeLibException(msg)

        if vs30grid.getData().dtype != np.float64:
            vs30grid.setData(vs30grid.getData().astype(np.float64))

        return vs30grid

    @staticmethod
    def _getFileGeoDict(fname):
        geodict = None
        try:
            geodict = get_file_geodict(fname)
        except Exception as msg1:
            msg = f'File geodict failure with {fname} - error messages: "{str(msg1)}"'
            raise ShakeLibException(msg)
        return geodict


def addDepthParameters(sctx):
    """
    Add the different depth parameters to a sites context from
    Vs30 values.

    Args:
        sctx: A sites context.

    Returns: A sites context with the depth parameters set.
    """
    sctx.z1pt0_cy14_cal = _z1pt0_from_vs30_cy14_cal(sctx.vs30)
    sctx.z1pt0_cy14_jpn = _z1pt0_from_vs30_cy14_jpn(sctx.vs30)
    sctx.z1pt0_ask14_cal = _z1pt0_from_vs30_ask14_cal(sctx.vs30)
    sctx.z1pt0_ask14_jpn = _z1pt0_from_vs30_ask14_jpn(sctx.vs30)
    sctx.z2pt5_cb14_cal = _z2pt5_from_vs30_cb14_cal(sctx.vs30)
    sctx.z2pt5_cb14_jpn = _z2pt5_from_vs30_cb14_jpn(sctx.vs30)
    sctx.z1pt0_cy08 = _z1pt0_from_vs30_cy08(sctx.vs30)
    sctx.z2pt5_cb07 = _z2pt5_from_z1pt0_cb07(sctx.z1pt0_cy08)


def _z1pt0_from_vs30_cy14_cal(vs30):
    """
    Compute z1.0 using CY14 relationship for California.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z1.0 in m.
    """
    z1 = np.exp(-(7.15 / 4.0) * np.log((vs30**4.0 + 571.0**4) / (1360**4.0 + 571.0**4)))
    return z1


def _z1pt0_from_vs30_cy14_jpn(vs30):
    """
    Compute z1.0 using CY14 relationship for Japan.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z1.0 in m.
    """
    z1 = np.exp(
        -(5.23 / 2.0) * np.log((vs30**2.0 + 412.0**2.0) / (1360**2.0 + 412.0**2.0))
    )
    return z1


def _z1pt0_from_vs30_ask14_cal(vs30):
    """
    Calculate z1.0 using ASK14 relationship for California.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z1.0 in m.

    """
    # ASK14 define units as km, but implemented as m in OQ
    z1 = np.exp(-(7.67 / 4.0) * np.log((vs30**4.0 + 610.0**4) / (1360**4.0 + 610.0**4)))
    return z1


def _z1pt0_from_vs30_ask14_jpn(vs30):
    """
    Calculate z1.0 using ASK14 relationship for Japan.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z1.0 in m.

    """
    # ASK14 define units as km, but implemented as m in OQ
    z1 = np.exp(-(5.32 / 2.0) * np.log((vs30**2.0 + 412.0**2) / (1360**2.0 + 412.0**2)))
    return z1


def _z2pt5_from_vs30_cb14_cal(vs30):
    """
    Calculate z2.5 using CB14 relationship for California.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z2.5 in m. *NOTE*: OQ's CampbellBozorgnia2014
        class expects z2.5 to be in km!
    """
    z2p5 = np.exp(7.089 - 1.144 * np.log(vs30))
    return z2p5


def _z2pt5_from_vs30_cb14_jpn(vs30):
    """
    Calculate z2.5 using CB14 relationship for Japan.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z2.5 in m. *NOTE*: OQ's CampbellBozorgnia2014
        class expects z2.5 to be in km!
    """
    z2p5 = np.exp(5.359 - 1.102 * np.log(vs30))
    return z2p5


def _z1pt0_from_vs30_cy08(vs30):
    """
    Chiou and Youngs (2008) z1.0 equation.

    Args:
        vs30: Numpy array of Vs30 values in m/s.

    Returns: Numpy array of z1.0 in m.
    """
    z1pt0 = np.exp(28.5 - (3.82 / 8.0) * np.log(vs30**8 + 378.7**8))
    return z1pt0


def _z2pt5_from_z1pt0_cb07(z1pt0):
    """
    Equations are from 2007 PEER report by Campbell and Bozorgnia.

    Args:
        z1pt0: Numpy array of z1.0 in m.

    Returns: Numpy array of z2.5 in m.
    """
    z2pt5 = 519.0 + z1pt0 * 3.595
    return z2pt5



#!/usr/bin/env python

# stdlib modules
import os.path
import sys

# third party modules
import numpy as np
import openquake.hazardlib.const as oqconst
import pytest

# local imports
from esi_shakelib.gmice.wgrw12 import WGRW12
from esi_shakelib.multigmpe import MultiGMPE
from esi_shakelib.sites import Sites
from esi_shakelib.utils.exception import ShakeLibException
from esi_shakelib.virtualipe import VirtualIPE
from esi_utils_rupture.distance import Distance
from esi_utils_rupture.factory import get_rupture
from esi_utils_rupture.origin import Origin
from openquake.hazardlib.gsim.boore_2014 import BooreEtAl2014
from openquake.hazardlib.gsim.chiou_youngs_2014 import ChiouYoungs2014
from openquake.hazardlib.imt import MMI, PGA, PGV, SA

homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, ".."))
sys.path.insert(0, shakedir)


def test_virtualipe():
    #
    # Set up the GMPE, IPE, and GMICE
    #
    gmpe_cy14 = ChiouYoungs2014()
    gmpe = MultiGMPE.__from_list__([gmpe_cy14], [1.0])
    gmice = WGRW12()
    ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)

    #
    # Use the Calexico event info
    #
    homedir = os.path.dirname(os.path.abspath(__file__))
    datadir = os.path.abspath(
        os.path.join(homedir, "virtualipe_data", "Calexico", "input")
    )

    #
    # Read the event, origin, and rupture files and produce Rupture and Origin
    # objects
    #
    # inputfile = os.path.join(datadir, 'stationlist_dat.xml')
    # dyfifile = os.path.join(datadir, 'ciim3_dat.xml')
    eventfile = os.path.join(datadir, "event.xml")
    rupturefile = os.path.join(datadir, "wei_fault.txt")

    origin_obj = Origin.fromFile(eventfile)
    rupture_obj = get_rupture(origin_obj, rupturefile)
    rx = rupture_obj.getRuptureContext([gmpe])
    rx.rake = 45.0

    smdx = 0.0083333333
    smdy = 0.0083333333
    lonspan = 6.0
    latspan = 4.0
    vs30filename = os.path.join(datadir, "..", "vs30", "vs30.grd")

    sites_obj_grid = Sites.fromCenter(
        rx.hypo_lon,
        rx.hypo_lat,
        lonspan,
        latspan,
        smdx,
        smdy,
        defaultVs30=760.0,
        vs30File=vs30filename,
        vs30measured_grid=None,
        padding=False,
        resample=False,
    )

    npts = 200
    lats = np.empty(npts)
    lons = np.empty(npts)
    depths = np.zeros(npts)
    for i in range(npts):
        lats[i] = rx.hypo_lat
        lons[i] = rx.hypo_lon + i * 0.01
    lldict = {"lats": lats, "lons": lons}

    sx = sites_obj_grid.getSitesContext(lldict=lldict, rock_vs30=760.0)

    dobj = Distance(gmpe, lons, lats, depths, rupture_obj)
    dx = dobj.getDistanceContext()

    sd_types = [oqconst.StdDev.TOTAL]
    mmi_const_vs30, mmi_sd_const_vs30 = ipe.get_mean_and_stddevs(
        sx, rx, dx, MMI(), sd_types
    )

    # These prints are just so a human can examine the outputs
    #    print(mmi_const_vs30)
    #    print(mmi_sd_const_vs30)

    sx = sites_obj_grid.getSitesContext(lldict=lldict)
    mmi_variable_vs30, mmi_sd_variable_vs30 = ipe.get_mean_and_stddevs(
        sx, rx, dx, MMI(), sd_types
    )

    #    print(mmi_variable_vs30)
    #    print(mmi_sd_variable_vs30)

    sd_types = [
        oqconst.StdDev.TOTAL,
        oqconst.StdDev.INTRA_EVENT,
        oqconst.StdDev.INTER_EVENT,
    ]
    mmi_variable_vs30_intra, mmi_sd_variable_vs30_intra = ipe.get_mean_and_stddevs(
        sx, rx, dx, MMI(), sd_types
    )

    #    print(mmi_variable_vs30_intra)
    #    print(mmi_sd_variable_vs30_intra)
    #    assert(0)      # Assert causes test to fail and prints to be displayed

    #
    # Try with PGA
    #
    gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES.remove(PGV)
    gmpe.ALL_GMPES_HAVE_PGV = False
    ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)
    mmi_pga, mmi_sd_pga = ipe.get_mean_and_stddevs(sx, rx, dx, MMI(), sd_types)
    #
    # Try with SA(1.0)
    #
    gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES.remove(PGA)
    ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)
    mmi_psa, mmi_sd_psa = ipe.get_mean_and_stddevs(sx, rx, dx, MMI(), sd_types)

    #
    # This should raise an exception because the IMT isn't MMI
    #
    with pytest.raises(ValueError) as e:
        mmi_psa, mmi_sd_psa = ipe.get_mean_and_stddevs(sx, rx, dx, PGA(), sd_types)
    #
    # This should raise an exception because no valid IMTs are available
    #
    gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES.remove(SA)
    with pytest.raises(ShakeLibException) as e:  # noqa
        ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)

    #
    # Now do a GMPE that uses Rjb instead of Rrup
    #
    gmpe_ba14 = BooreEtAl2014()
    gmpe = MultiGMPE.__from_list__([gmpe_ba14], [1.0])
    ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)
    rx = rupture_obj.getRuptureContext([gmpe])
    rx.rake = 45.0
    dobj = Distance(gmpe, lons, lats, depths, rupture_obj)
    dx = dobj.getDistanceContext()

    mmi_rjb, mmi_sd_rjb = ipe.get_mean_and_stddevs(sx, rx, dx, MMI(), sd_types)

    #
    # Test the results against a known standard
    #
    savefile = os.path.abspath(
        os.path.join(
            homedir, "virtualipe_data", "Calexico", "virtualipe_test", "savefile.npz"
        )
    )

    #
    # If things change, set remake_save to True, and it will rebuild the
    # saved data file against which the comparisons are done
    # Remember to set this back to False once you've remade the test datafile
    #
    remake_save = False
    if remake_save:
        np.savez_compressed(
            savefile,
            mmi_const_vs30=mmi_const_vs30,
            mmi_sd_const_vs30=mmi_sd_const_vs30[0],
            mmi_variable_vs30=mmi_variable_vs30,
            mmi_sd_variable_vs30=mmi_sd_variable_vs30[0],
            mmi_variable_vs30_intra=mmi_variable_vs30_intra,
            mmi_sd_variable_vs30_total=mmi_sd_variable_vs30_intra[0],
            mmi_sd_variable_vs30_intra=mmi_sd_variable_vs30_intra[1],
            mmi_sd_variable_vs30_inter=mmi_sd_variable_vs30_intra[2],
            mmi_pga=mmi_pga,
            mmi_sd_pga=mmi_sd_pga[0],
            mmi_psa=mmi_psa,
            mmi_sd_psa=mmi_sd_psa[0],
            mmi_rjb=mmi_rjb,
            mmi_sd_rjb=mmi_sd_rjb[0],
        )

    td = np.load(savefile)

    assert np.allclose(td["mmi_const_vs30"], mmi_const_vs30)
    assert np.allclose(td["mmi_sd_const_vs30"], mmi_sd_const_vs30[0])
    assert np.allclose(td["mmi_variable_vs30"], mmi_variable_vs30)
    assert np.allclose(td["mmi_sd_variable_vs30"], mmi_sd_variable_vs30[0])
    assert np.allclose(td["mmi_variable_vs30_intra"], mmi_variable_vs30_intra)
    assert np.allclose(td["mmi_sd_variable_vs30_total"], mmi_sd_variable_vs30_intra[0])
    assert np.allclose(td["mmi_sd_variable_vs30_intra"], mmi_sd_variable_vs30_intra[1])
    assert np.allclose(td["mmi_sd_variable_vs30_inter"], mmi_sd_variable_vs30_intra[2])
    assert np.allclose(td["mmi_pga"], mmi_pga)
    assert np.allclose(td["mmi_sd_pga"], mmi_sd_pga[0])
    assert np.allclose(td["mmi_psa"], mmi_psa)
    assert np.allclose(td["mmi_sd_psa"], mmi_sd_psa[0])
    assert np.allclose(td["mmi_rjb"], mmi_rjb)
    assert np.allclose(td["mmi_sd_rjb"], mmi_sd_rjb[0])

    # The total uncertainties should be greater than the intra-event
    assert np.all(mmi_sd_variable_vs30[0] > mmi_sd_variable_vs30_intra[1])

    # The combined intra and inter-event uncertainty should be equal
    # to the total
    tot = np.sqrt(
        mmi_sd_variable_vs30_intra[1] ** 2 + mmi_sd_variable_vs30_intra[2] ** 2
    )
    assert np.allclose(tot, mmi_sd_variable_vs30_intra[0], rtol=1e-2)

    # Run through mgm = mgm + fd if fd is available
    # In this case fd = 0, so the ouput is still the same values
    mmi_fd, mmi_fdsd = ipe.get_mean_and_stddevs(sx, rx, dx, MMI(), sd_types, fd=0)
    assert np.allclose(mmi_fd, mmi_rjb)
    assert np.allclose(mmi_fdsd, mmi_sd_rjb)


if __name__ == "__main__":
    test_virtualipe()



"""
Module to implement CB14 basin and M9 modifications to Atkinson and Macias (2009) GSIM.
"""

from esi_shakelib.gmpe.modifiable_gmpes.ak_inter_bias import NSHM_AK_Bias_Correction


class ParkerEtAl2020SInterAK(NSHM_AK_Bias_Correction):
    """
    Returns bias adjustments to the Parker et al. (2020) interface GSIM
    """

    def __init__(self):
        super().__init__(gmpe_name="ParkerEtAl2020SInter")


class AbrahamsonGulerce2020SInterAK(NSHM_AK_Bias_Correction):
    """
    Returns bias adjustments to the Abrahamson and Gulerce (2020) interface GSIM
    """

    def __init__(self):
        super().__init__(gmpe_name="AbrahamsonGulerce2020SInter")


class KuehnEtAl2020SInterAK(NSHM_AK_Bias_Correction):
    """
    Returns bias adjustments to the Kuehn et al. (2020) interface GSIM
    """

    def __init__(self):
        super().__init__(gmpe_name="KuehnEtAl2020SInter")



# stdlib imports
import logging
import importlib.metadata

# third party imports
import numpy as np
from strec.gmreg import Regionalizer

from esi_utils_rupture.base import Rupture
from esi_utils_rupture.edge_rupture import EdgeRupture
from esi_utils_rupture.quad_rupture import QuadRupture
from openquake.hazardlib import const, imt
from openquake.hazardlib.geo.utils import OrthographicProjection
from esi_shakelib.utils.contexts import (
    SitesContext,
    DistancesContext,
    RuptureContext,
)
from esi_shakelib.multigmpe import set_sites_depth_parameters
from esi_shakelib.station import StationList

DEFAULT_ACTIVE_COEFFS = [27.24, 250.4, 579.1]
DEFAULT_STABLE_COEFFS = [63.4, 465.4, 581.3]


def replace_dyfi(stationfile, dyfi_xml, min_nresp=3):
    """Remove any non-instrumented data from station file, add DYFI.

    Args:
        stationfile (str): Existing station data file, presumed to
                           contain old DYFI data.
        dyfi_xml (str): DYFI XML data file, which will be added to
                        instrumented station data.
        min_nresp (int): The minimum number of DYFI responses required for an
                         observation to be added to the stationlist.
                         Default = 3.
    Returns:
        StationList: Object containing merged data.

    """
    stations = StationList.loadFromFiles([stationfile], min_nresp=1)
    # reach into the internal database and find the instrumented stations
    conn = stations.db
    cursor = stations.cursor
    query1 = (
        "SELECT id from station WHERE instrumented = "
        "0 and (network = 'DYFI' or network='CIIM')"
    )
    cursor.execute(query1)
    rows = cursor.fetchall()
    for row in rows:
        sid = row[0]
        query2 = f"DELETE FROM amp WHERE station_id='{sid}'"
        cursor.execute(query2)
        conn.commit()
        query3 = f"DELETE FROM station where id='{sid}'"
        cursor.execute(query3)
        conn.commit()

    # now insert the dyfi data
    stations.addData([dyfi_xml], min_nresp)
    return stations


def get_extent(config, ipe, rupture=None):
    """
    Method to compute map extent from rupture. There are numerous methods for
    getting the extent:
        - It can be specified directly in the config file,
        - it can be hard coded for specific magnitude ranges in the config
          file, or
        - it can be based on the MultiGMPE for the event.

    All methods except for the first require a rupture object.

    If no config is provided then a rupture is required and the extent is based
    on a generic set of active/stable.

    Args:
        config (ConfigObj):
            ShakeMap config object.
        ipe (VirtualIPE):
            An VirtualIPE instance.
        rupture (Rupture):
            A ShakeMap Rupture instance.

    Returns:
        tuple: lonmin, lonmax, latmin, latmax rounded outward to the nearest
        30 arc seconds.

    """

    # -------------------------------------------------------------------------
    # Check to see what parameters are specified in the extent config
    # -------------------------------------------------------------------------
    spans = {}
    bounds = []
    offsets = None
    if "extent" in config:
        if "magnitude_spans" in config["extent"]:
            if len(config["extent"]["magnitude_spans"]):
                if isinstance(config["extent"]["magnitude_spans"], dict):
                    spans = config["extent"]["magnitude_spans"]
        if "bounds" in config["extent"]:
            if "extent" in config["extent"]["bounds"]:
                if config["extent"]["bounds"]["extent"][0] != -999.0:
                    bounds = config["extent"]["bounds"]["extent"]
        if "relative_offset" in config["extent"]:
            if isinstance(config["extent"]["relative_offset"], list):
                offsets = config["extent"]["relative_offset"]

    # -------------------------------------------------------------------------
    # Simplest option: extent was specified in the config, use that and exit.
    # -------------------------------------------------------------------------
    if len(bounds) > 0:
        xmin, ymin, xmax, ymax = bounds
        return (
            thirty_sec_min(xmin),
            thirty_sec_max(xmax),
            thirty_sec_min(ymin),
            thirty_sec_max(ymax),
        )

    if not rupture or not isinstance(rupture, Rupture):
        raise TypeError(
            "get_extent() requires a rupture object if the extent "
            "is not specified in the config object."
        )

    # -------------------------------------------------------------------------
    # Second simplest option: spans are hardcoded based on magnitude
    # -------------------------------------------------------------------------
    extent = None
    if len(spans):
        extent = _get_extent_from_spans(rupture, spans)

    # -------------------------------------------------------------------------
    # Otherwise, use MultiGMPE to get spans
    # -------------------------------------------------------------------------
    if extent is None:
        extent = _get_extent_from_multigmpe(rupture, config, ipe)

    if offsets is None:
        xmin, xmax, ymin, ymax = extent
        return (
            thirty_sec_min(xmin),
            thirty_sec_max(xmax),
            thirty_sec_min(ymin),
            thirty_sec_max(ymax),
        )

    # -------------------------------------------------------------------------
    # Apply relative offsets
    # -------------------------------------------------------------------------
    (xmin, xmax, ymin, ymax) = extent

    xspan = xmax - xmin
    yspan = ymax - ymin

    xmin += xspan * offsets[0]
    xmax += xspan * offsets[0]
    ymin += yspan * offsets[1]
    ymax += yspan * offsets[1]

    return (
        thirty_sec_min(xmin),
        thirty_sec_max(xmax),
        thirty_sec_min(ymin),
        thirty_sec_max(ymax),
    )


def _get_extent_from_spans(rupture, spans={}):
    """
    Choose extent based on magnitude using a hardcoded list of spans
    based on magnitude ranges.
    """
    (clon, clat) = _rupture_center(rupture)
    mag = rupture.getOrigin().mag
    xmin = None
    xmax = None
    ymin = None
    ymax = None
    for span in spans.values():
        if span[0] < mag <= span[1]:
            ymin = clat - span[2] / 2
            ymax = clat + span[2] / 2
            xmin = clon - span[3] / 2
            xmax = clon + span[3] / 2
            break
    if xmin is not None:
        return (xmin, xmax, ymin, ymax)
    return None


def _get_extent_from_multigmpe(rupture, config, ipe):
    """
    Use MultiGMPE to determine extent

    Args:
        rupture (Rupture):
            A ShakeMap Rupture instance.
        config (ConfigObj):
            ShakeMap config object.
        ipe (VirtualIPE):
            An VirtualIPE instance.
    """
    (clon, clat) = _rupture_center(rupture)
    origin = rupture.getOrigin()

    min_mmi = config["extent"]["mmi"]["threshold"]
    sd_types = [const.StdDev.TOTAL]

    # Distance context
    dx = DistancesContext()
    size = 200
    # This imposes minimum/ maximum distances of:
    #   80 and 800 km; could make this configurable
    d_min = config["extent"]["mmi"]["mindist"]
    d_max = config["extent"]["mmi"]["maxdist"]
    dx.rjb = np.logspace(np.log10(d_min), np.log10(d_max), size)
    dx.rrup = np.sqrt(dx.rjb**2 + origin.depth**2)
    dx.rhypo = dx.rrup
    dx.repi = dx.rjb
    dx.rx = np.zeros_like(dx.rjb)
    dx.ry0 = np.zeros_like(dx.rjb)
    dx.rvolc = np.zeros_like(dx.rjb)

    # Sites context
    sx = SitesContext
    # Set to soft soil conditions
    sx.sids = np.array(range(size))
    sx.vs30 = np.full(size, 180.0)
    sx.lons = np.full(size, 0.0)
    set_sites_depth_parameters(sx, ipe)
    sx.vs30measured = np.full(size, False, dtype=bool)
    sx.backarc = np.full(size, False, dtype=bool)

    # Rupture context
    rx = RuptureContext()
    rx.mag = np.full(size, origin.mag)
    rx.rake = np.zeros(size)
    # From WC94...
    # rx.width = 10 ** (-0.76 + 0.27 * rx.mag)
    # rx.dip = 90.0
    # rx.ztor = origin.depth
    # Using parameters from the point rupture...
    rx.width = np.full(size, rupture.getWidth())
    rx.dip = np.full(size, rupture.getDip())
    rx.ztor = np.full(size, rupture.getDepthToTop())
    rx.hypo_depth = np.full(size, origin.depth)

    mmi = imt.from_string("MMI")
    imt_mean, _ = ipe.get_mean_and_stddevs(sx, rx, dx, mmi, sd_types)

    # Minimum distance that exceeds threshold MMI?
    dists_exceed_mmi = dx.rjb[imt_mean[0] > min_mmi]
    if len(dists_exceed_mmi):
        mindist_km = np.max(dists_exceed_mmi)
    else:
        mindist_km = d_min

    # Get a projection
    proj = OrthographicProjection(clon - 4, clon + 4, clat + 4, clat - 4)
    if isinstance(rupture, (QuadRupture, EdgeRupture)):
        ruptx, rupty = proj(
            rupture.lons[~np.isnan(rupture.lons)],
            rupture.lats[~np.isnan(rupture.lats)],
        )
    else:
        ruptx, rupty = proj(clon, clat)

    xmin = np.nanmin(ruptx) - mindist_km
    ymin = np.nanmin(rupty) - mindist_km
    xmax = np.nanmax(ruptx) + mindist_km
    ymax = np.nanmax(rupty) + mindist_km

    # Put a limit on range of aspect ratio
    dx = xmax - xmin
    dy = ymax - ymin
    ar = dy / dx
    if ar > 1.2:
        # Inflate x
        dx_target = dy / 1.2
        ddx = dx_target - dx
        xmax = xmax + ddx / 2
        xmin = xmin - ddx / 2
    if ar < 0.83:
        # Inflate y
        dy_target = dx * 0.83
        ddy = dy_target - dy
        ymax = ymax + ddy / 2
        ymin = ymin - ddy / 2

    lonmin, latmin = proj(np.array([xmin]), np.array([ymin]), reverse=True)
    lonmax, latmax = proj(np.array([xmax]), np.array([ymax]), reverse=True)

    #
    # Round coordinates to the nearest minute -- that should make the
    # output grid register with common grid resolutions (60c, 30c,
    # 15c, 7.5c)
    #
    logging.debug(
        "Extent: %f, %f, %f, %f", lonmin[0], lonmax[0], latmin[0], latmax[0]
    )
    return (
        _round_coord(lonmin[0]),
        _round_coord(lonmax[0]),
        _round_coord(latmin[0]),
        _round_coord(latmax[0]),
    )


def _rupture_center(rupture):
    """
    Find the central point of a rupture
    """
    origin = rupture.getOrigin()
    if isinstance(rupture, (QuadRupture, EdgeRupture)):
        # For an extended rupture, it is the midpoint between the extent of the
        # verticies
        lats = rupture.lats
        lons = rupture.lons

        # Remove nans
        lons = lons[~np.isnan(lons)]
        lats = lats[~np.isnan(lats)]

        clat = 0.5 * (np.nanmax(lats) + np.nanmin(lats))
        clon = 0.5 * (np.nanmax(lons) + np.nanmin(lons))
    else:
        # For a point source, it is just the epicenter
        clat = origin.lat
        clon = origin.lon
    return (clon, clat)


def _round_coord(coord):
    """
    Round a number to the nearest arc-minute
    """
    dm = 1.0 / 60
    mm = coord / dm
    imm = int(mm + 0.5)
    return imm * dm


def thirty_sec_min(coord):
    """
    Round a number to the floor of 30 arc-seconds
    """
    return np.floor(coord * 120.0) / 120.0


def thirty_sec_max(coord):
    """
    Round a number to the ceiling of 30 arc-seconds
    """
    return np.ceil(coord * 120.0) / 120.0


def is_stable(lon, lat):
    """
    Determine if point is located in the US stable tectonic region. This uses
    STREC but only makes use of the stable tectonic region. Any location that
    is not mapped as stable is classified as active.

    Args:
        lon (float): Lognitude.
        lat (float): Latitude.
    Returns:
        bool: Is the point classified as tectonically stable.
    """
    reg = Regionalizer.load()
    region_info = reg.getRegions(lat, lon)
    return region_info["TectonicRegion"] == "Stable"


def get_shakelib_version():
    """
    Get the code version of this package.
    """
    return importlib.metadata.version("esi_shakelib")


def get_gmpe_coef_table(gmpe):
    """
    Method for finding the (or "a") GMPE table.

    Notes:

      *  The reason for the complexity here is that there can be multiple
         coefficient tables, and some of them may not have the sa_coeffs
         attribute, which is the main reason for getting the table.
      *  We are also assuming that if there are more than one  coefficient
         table, the range of periods will be the same across all of the
         tables.

    Args:
        gmpe (GMPE): An OQ GMPE instance.

    Returns:
        The associated coefficient table.

    """
    stuff = dir(gmpe)
    coef_list = [s for s in stuff if "COEFFS" in s]
    for coef_sel in coef_list:
        cobj = getattr(gmpe, coef_sel)
        if "sa_coeffs" in dir(cobj):
            return cobj
    raise NotImplementedError(
        f"GMPE {gmpe} does not contain sa_coeffs attribute."
    )


def get_gmpe_sa_periods(gmpe):
    """
    Method to extract the SA periods defined by a GMPE.

    Args:
        gmpe (GMPE): A GMPE instance.

    Retunrs:
        list: List of periods.

    """
    if gmpe == "[NGAEast]":
        per = gmpe.per_array
    else:
        ctab = get_gmpe_coef_table(gmpe).sa_coeffs
        ilist = list(ctab.keys())
        per = [i.period for i in ilist]
    return per



"""
Module implements BeyerBommer2006 class to convert between various
horizontal intensity measure components.
"""
import logging

# Third party imports
from openquake.hazardlib import const
from openquake.hazardlib.const import IMC
import numpy as np

# Local imports
from esi_shakelib.conversions.convert_imc import ComponentConverter


class BeyerBommer2006(ComponentConverter):
    """
    Implements conversion for various "Intensity Measure
    Components" (IMCs) per Beyer and Bommer (2006).

    IMC equivalencies:

    +---------------------------+------------------------+
    | OpenQuake                 | Beyer & Bommer         |
    +===========================+========================+
    | GEOMETRIC_MEAN            | GMxy (Geometric mean)  |
    +---------------------------+------------------------+
    | HORIZONTAL                | Random?                |
    +---------------------------+------------------------+
    | MEDIAN_HORIZONTAL         | AMxy (Arithmetic mean) |
    +---------------------------+------------------------+
    | GMRotI50                  | GMRotI50               |
    +---------------------------+------------------------+
    | RotD50                    | GMRotD50               |
    +---------------------------+------------------------+
    | RANDOM_HORIZONTAL         | Random                 |
    +---------------------------+------------------------+
    | GREATER_OF_TWO_HORIZONTAL | Env_xy                 |
    +---------------------------+------------------------+
    | VERTICAL                  | ---                    |
    +---------------------------+------------------------+

    Notes

        - GEOMETRIC_MEAN is the "reference" type.
        - The OQ IMC "HORIZONAL" indicates that the horizontal IMC category
          may be ambiguous. In these cases, we are assuming that it is a
          random horizontal component as a default.
        - Assumes ALL unknown IMC types are GEOMETRIC_MEAN

    References

        Beyer, K., & Bommer, J. J. (2006). Relationships between median values
        and between aleatory variabilities for different definitions of the
        horizontal component of motion. Bulletin of the Seismological Society
        of America, 96(4A), 1512-1522.
        `[link] <http://www.bssaonline.org/content/96/4A/1512.short>`__

    """

    def __init__(self, imc_in, imc_out):
        super().__init__()
        if type(imc_in) == IMC:
            self.imc_in = imc_in
        elif type(imc_in) == str:
            imc_from_str = self.imc_from_str(imc_in)
            if imc_from_str is None:
                logging.warning(
                    "Unknown IMC string '%s' in input, using Geometric Mean",
                    imc_in,
                )
                self.imc_in = IMC.GEOMETRIC_MEAN
            else:
                self.imc_in = imc_from_str
        else:
            logging.warning(
                "Unknown object passed as input %s, using Geometric Mean",
                type(imc_in),
            )
            self.imc_in = IMC.GEOMETRIC_MEAN
        if type(imc_out) == IMC:
            self.imc_out = imc_out
        elif type(imc_out) == str:
            imc_from_str = self.imc_from_str(imc_out)
            if imc_from_str is None:
                logging.warning(
                    "Unknown IMC string '%s' in output, using Geometric Mean",
                    imc_out,
                )
                self.imc_out = IMC.GEOMETRIC_MEAN
            else:
                self.imc_out = imc_from_str
        else:
            logging.warning(
                "Unknown object passed as opuput %s, using Geometric Mean",
                type(imc_out),
            )
            self.imc_out = IMC.GEOMETRIC_MEAN
        # c12 = median ratio
        # c34 = std. of log ratio
        # R = ratio of sigma_pga values
        self.__pga_pgv_col_names = ["c12", "c34", "R"]
        self.__sa_col_names = ["c1", "c2", "c3", "c4", "R"]
        # Possible conversions
        self.conversion_graph = {
            IMC.GREATER_OF_TWO_HORIZONTAL: set(
                [
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RotD50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.MEDIAN_HORIZONTAL: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RotD50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.GMRotI50: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.RotD50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.RotD50: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.RANDOM_HORIZONTAL: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RotD50,
                    IMC.HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.HORIZONTAL: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RotD50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.GEOMETRIC_MEAN,
                ]
            ),
            IMC.GEOMETRIC_MEAN: set(
                [
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                    IMC.GMRotI50,
                    IMC.RotD50,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                ]
            ),
        }
        # Check if any imc values are unknown. If they are, convert
        # to GEOMETRIC_MEAN
        self.checkUnknown()
        # Get shortest conversion "path" between imc_in and imc_out
        self.path = self.getShortestPath(
            self.conversion_graph, self.imc_in, self.imc_out
        )
        # Set dictionaries of constants
        self.__pga_dict = {
            const.IMC.MEDIAN_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.01, 1.00]))
            ),
            const.IMC.GMRotI50: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.02, 1.00]))
            ),
            const.IMC.RotD50: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.02, 1.00]))
            ),
            const.IMC.RANDOM_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.07, 1.03]))
            ),
            const.IMC.HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.07, 1.03]))
            ),
            const.IMC.GREATER_OF_TWO_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.1, 0.05, 1.02]))
            ),
        }
        self.__pgv_dict = {
            const.IMC.MEDIAN_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.01, 1.00]))
            ),
            const.IMC.GMRotI50: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.02, 1.00]))
            ),
            const.IMC.RotD50: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.02, 1.00]))
            ),
            const.IMC.RANDOM_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.07, 1.03]))
            ),
            const.IMC.HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.0, 0.07, 1.03]))
            ),
            const.IMC.GREATER_OF_TWO_HORIZONTAL: dict(
                list(zip(self.__pga_pgv_col_names, [1.1, 0.05, 1.02]))
            ),
        }
        self.__sa_dict = {
            const.IMC.MEDIAN_HORIZONTAL: dict(
                list(zip(self.__sa_col_names, [1.0, 1.0, 0.01, 0.02, 1.00]))
            ),
            const.IMC.GMRotI50: dict(
                list(zip(self.__sa_col_names, [1.0, 1.0, 0.03, 0.04, 1.00]))
            ),
            const.IMC.RotD50: dict(
                list(zip(self.__sa_col_names, [1.0, 1.0, 0.02, 0.03, 1.00]))
            ),
            const.IMC.RANDOM_HORIZONTAL: dict(
                list(zip(self.__sa_col_names, [1.0, 1.0, 0.07, 0.11, 1.05]))
            ),
            const.IMC.HORIZONTAL: dict(
                list(zip(self.__sa_col_names, [1.0, 1.0, 0.07, 0.11, 1.05]))
            ),
            const.IMC.GREATER_OF_TWO_HORIZONTAL: dict(
                list(zip(self.__sa_col_names, [1.1, 1.2, 0.04, 0.07, 1.02]))
            ),
        }

    def convertAmpsOnce(self, imt, amps, rrups=None, mag=None):
        """
        Returns amps converted from one IMC to another.

        **Important**:

            - Assumes the input amps are in natural log (not linear) space
            - IMC type 'VERTICAL' is not supported

        Args:
            imt (IMT): OpenQuake IMT of the input amps (must be one of PGA,
                PGV, or SA).
                `[link] <http://docs.openquake.org/oq-hazardlib/master/imt.html>`

            amps (array): Numpy array of ground motion amplitudes.
            rrups (array): A numpy array of the same shape as amps,
                containing the rupture distances of the ground motions.
                Ignored by this method.
            mag (float): The earthquake magnitude. Default is None.
                 Ignored by this method.

        Returns:
            array: Numpy array of amps converted from imc_in to imc_out.
        """

        denom = self.__GM2other(imt, self.imc_in)
        numer = self.__GM2other(imt, self.imc_out)

        return amps + np.log(numer / denom)

    def convertSigmasOnce(self, imt, sigmas):
        """
        Returns standard deviations converted from one IMC to another.

        **Important**:

            - Assumes the input sigmas are in natural log space
            - IMC types 'VERTICAL' and 'HORIZONTAL' are not supported

        Args:
            imt (IMT): OpenQuake IMT of the input sigmas (must be one of PGA,
                 PGV, or SA) `[link] <http://docs.openquake.org/oq-hazardlib/master/imt.html>`__
            sigmas (array): Numpy array of standard deviations.

        Returns:
            array: Numpy array of standard deviations converted from imc_in to
                imc_out.

        """
        # ---------------------------------------------------------------------
        # Take input sigma to geometric mean sigma.
        # Solve eqn 8 for sigma_GM2, which is sigma_logSa_GM
        # This is the sigma converted to geometric mean
        # (i.e., reference component).
        # ---------------------------------------------------------------------
        R, sig_log_ratio = self.__GM2otherSigma(imt, self.imc_in)
        sigma_GM2 = (sigmas ** 2 - sig_log_ratio ** 2) / R ** 2

        # ---------------------------------------------------------------------
        # Evaluate equation 8 to go from GM to requested IMC
        # ---------------------------------------------------------------------
        R, sig_log_ratio = self.__GM2otherSigma(imt, self.imc_out)
        sigma_out = np.sqrt(sigma_GM2 * R ** 2 + sig_log_ratio ** 2)

        return sigma_out

    def __GM2other(self, imt, imc):
        """
        Helper function to extract coefficients from the parameters for
        converting the median ground motions.

        Args:
            imt (IMT): OQ intensity measure type.
            imc (IMC): OQ Intensity measure component.

        Returns:
            float: Median ratios. This is directly from Table 2 for PGA and
                PGV, and computed from coefficients in Table 3 along with
                eqn 10 for Sa.

        Raises:
            ValueError if unknown IMT specified.
        """
        if imc == const.IMC.GEOMETRIC_MEAN:
            # The amps are already in the B&B "reference" type ("GM", i.e.,
            # geometric mean)
            return 1.0

        self._verifyConversion(imc)

        if "PGA" in imt.string:
            return self.__pga_dict[imc]["c12"]
        elif "PGV" in imt.string:
            return self.__pgv_dict[imc]["c12"]
        elif "SA" in imt.string:
            pp = imt.period
            if pp <= 0.15:
                return self.__sa_dict[imc]["c1"]
            elif pp < 0.8:
                c1 = self.__sa_dict[imc]["c1"]
                c2 = self.__sa_dict[imc]["c2"]
                return c1 + (c2 - c1) * np.log(pp / 0.15) / np.log(0.8 / 0.15)
            elif pp <= 5.0:
                return self.__sa_dict[imc]["c2"]
            else:
                # Not sure what's right here; should probably raise an error
                # but for now let's just use c2
                return self.__sa_dict[imc]["c2"]
        else:
            raise ValueError(f"unknown IMT {imt.string}")

    def __GM2otherSigma(self, imt, imc):
        """
        Helper function to extract coefficients from the parameters for
        converting standard deviations.

        Args:
            imt (IMT): OQ intensity measure type.
            imc (IMC): OQ intensity measure component.

        Returns:
            tuple: Coefficients: R (ratio of sigma values), standard deviation
            of sigma ratios.

        Raises:
            ValueError if unknown IMT specified.
        """

        if imc == const.IMC.GEOMETRIC_MEAN:
            # The amps are already in the B&B "reference" type ("GM", i.e.,
            # geometric mean)
            return 1.0, 0.0

        self._verifyConversion(imc)

        if "PGA" in imt.string:
            return self.__pga_dict[imc]["R"], self.__pga_dict[imc]["c34"]
        elif "PGV" in imt.string:
            return self.__pgv_dict[imc]["R"], self.__pgv_dict[imc]["c34"]
        elif "SA" in imt.string:
            R = self.__sa_dict[imc]["R"]
            pp = imt.period
            if pp <= 0.15:
                return R, self.__sa_dict[imc]["c3"]
            elif pp < 0.8:
                c3 = self.__sa_dict[imc]["c3"]
                c4 = self.__sa_dict[imc]["c4"]
                return R, c3 + (c4 - c3) * np.log(pp / 0.15) / np.log(0.8 / 0.15)
            elif pp <= 5.0:
                return R, self.__sa_dict[imc]["c4"]
            else:
                # Not sure what's right here; should probably raise an error
                # but for now let's just use c4
                return R, self.__sa_dict[imc]["c4"]
        else:
            raise ValueError(f"unknown IMT {imt.string}")

    def _verifyConversion(self, imc_in, imc_out=None):
        """
        Helper method to ensure that the conversion is possible.

        Args:
            imc_in (IMC): OpenQuake IMC type of the input amp array.
            imc_out (IMC): Desired OpenQuake IMC type of the output amps.
                Default is None. Ignored by this method.

        Raises:
            ValueError if imc_in is not valid.
        """
        if (
            imc_in != const.IMC.GREATER_OF_TWO_HORIZONTAL
            and imc_in != const.IMC.MEDIAN_HORIZONTAL
            and imc_in != const.IMC.GMRotI50
            and imc_in != const.IMC.RotD50
            and imc_in != const.IMC.RANDOM_HORIZONTAL
            and imc_in != const.IMC.HORIZONTAL
        ):
            raise ValueError(f"unknown IMC {imc_in}")



import numpy as np

from openquake.hazardlib import const


class AbrahamsonBhasin2020(object):
    """
    Module to make conditional conversions from SA(T) to PGV. The user
    provides a magnitude when an instance and the instance can report back
    what period of SA is appropriate to use for computing PGV. The module
    is based upon:

    Abrahamson, N.A., and Bhasin, S. (2020) Conditional ground-motion model
    for peak ground velocity for active crustal regions. PEER Report 2020/05,
    Pacific Earthquake Engineering Research Center, U.C. Berkeley.
    """

    def __init__(self, mag):
        self.mag = mag
        mag_T = [
            (3.5, 0.20),
            (4.5, 0.28),
            (5.5, 0.40),
            (6.5, 0.95),
            (7.5, 1.40),
            (8.5, 2.80),
        ]
        self.Tref = 0
        for mm, tt in mag_T:
            if np.any(mag <= mm):
                self.Tref = tt
                break
        if self.Tref == 0:
            self.Tref = tt
        self.coeff = {
            "a1": 5.39,
            "a2": 0.799,
            "a3": 0.654,
            "a4": 0.479,
            "a5": -0.062,
            "a6": -0.359,
            "a7": -0.134,
            "a8": 0.023,
            "phi": 0.29,
            "tau": 0.16,
            "sigma": 0.33,
        }

    def getTref(self):
        """
        Returns the period of the SA amps suitable for conversion to PGV.
        Args:
            None.

        Returns:
            (float): The period of the SA amps to supply to getPGVandSTDDEVS()
        """
        return self.Tref

    def getPGVandSTDDEVS(self, psa, stddevs, sdtypes, rrup, vs30):
        """
        Function to compute the PGVs and inflate the uncertainties based on
        the input SA amps and uncertainties.

        Args:
            psa (np.array): An array os SA amp of the appropriate period.
            stddevs(np.arrays): A list of one or more array of standard
                deviations of the psa amps. The list may also contain arrays
                of the stddevs inflated by the point-source adjustments. Each
                array must be the same shape as the psa array.
            sdtypes (list): A list of the types of the standard deviations in
                stddevs (in order). These are OpenQuake const.StdDev types.
            rrup (np.array): Rupture distance to the points in psa. Must be
                the same shape as psa.
            vs30 (np.array): The Vs30 value at the points in psa. Must be the
                same shape as psa.

        Returns:
            (tuple: ndarray, list): An array of PGV values, and a list of
            stddev arrays corresponding to the input stddev types, now
            with the uncertainty propagated through the conversion.
        """
        c = self.coeff
        m = self.mag
        if np.any(m < 5):
            f1 = c["a2"]
        elif np.any(m <= 7.5):
            f1 = c["a2"] + (c["a3"] - c["a2"]) * (m - 5.0) / 2.5
        else:
            f1 = c["a3"]

        pgv = (
            c["a1"]
            + f1 * psa
            + c["a4"] * (m - 6.0)
            + c["a5"] * (8.5 - m) ** 2
            + c["a6"] * np.log(rrup + 5.0 * np.exp(0.4 * (m - 6.0)))
            + (c["a7"] + c["a8"] * (m - 5.0)) * np.log(vs30 / 425)
        )

        #
        # ShakeMap will often produce lists of stddevs that are twice
        # as long as sdtypes: one set that has been inflated by the
        # point-source to finite-fault factors, and one that has not.
        # Here we modify both, if they are present.
        #
        stddevs_out = []
        mytypes = sdtypes.copy()
        if len(stddevs) == 2 * len(sdtypes):
            mytypes.extend(sdtypes)
        for i, sdtype in enumerate(mytypes):
            if sdtype == const.StdDev.INTER_EVENT:
                sdout = np.sqrt(f1 ** 2 * stddevs[i] ** 2 + c["tau"] ** 2)
            elif sdtype == const.StdDev.INTRA_EVENT:
                sdout = np.sqrt(f1 ** 2 * stddevs[i] ** 2 + c["phi"] ** 2)
            else:
                sdout = np.sqrt(f1 ** 2 * stddevs[i] ** 2 + c["sigma"] ** 2)
            stddevs_out.append(sdout)

        return pgv, stddevs_out


class AbrahamsonBhasin2020PGA(AbrahamsonBhasin2020):
    """
    An alternative to AbrahamsonBhasin2020 when only PGA is available for
    conversion.
    """

    def __init__(self, mag):
        self.mag = mag
        self.Tref = 0
        self.coeff = {
            "a1": 4.77,
            "a2": 0.738,
            "a3": 0.484,
            "a4": 0.275,
            "a5": -0.036,
            "a6": -0.332,
            "a7": -0.44,
            "a8": 0.0,
            "phi1": 0.32,
            "phi2": 0.42,
            "phi": 0.0,
            "tau1": 0.12,
            "tau2": 0.26,
            "tau": 0.0,
            "sigma1": 0.34,
            "sigma2": 0.49,
            "sigma": 0.0,
            "m1": 5.0,
            "m2": 7.0,
        }
        c = self.coeff
        mref = (self.mag - c["m1"]) / (c["m2"] - c["m1"])

        #
        # Here we set the magnitude-dependent phi, tau, and sigma
        # to be used in the parent cless' getPGVandSTDDEVS()
        #
        if self.mag < c["m1"]:
            c["phi"] = c["phi1"]
            c["tau"] = c["tau1"]
            c["sigma"] = c["sigma1"]
        elif self.mag <= c["m2"]:
            c["phi"] = c["phi1"] + (c["phi2"] - c["phi2"]) * mref
            c["tau"] = c["tau1"] + (c["tau2"] - c["tau2"]) * mref
            c["sigma"] = c["sigma1"] + (c["sigma2"] - c["sigma2"]) * mref
        else:
            c["phi"] = c["phi2"]
            c["tau"] = c["tau2"]
            c["sigma"] = c["sigma2"]


class AbrahamsonBhasin2020SA1(AbrahamsonBhasin2020):
    """
    An alternative to AbrahamsonBhasin2020 when only SA(1.0) is available for
    conversion.
    """

    def __init__(self, mag):
        self.mag = mag
        self.Tref = 1.0
        self.coeff = {
            "a1": 4.80,
            "a2": 0.82,
            "a3": 0.55,
            "a4": 0.27,
            "a5": 0.054,
            "a6": -0.382,
            "a7": -0.21,
            "a8": 0.0,
            "phi1": 0.28,
            "phi2": 0.38,
            "phi": 0.0,
            "tau1": 0.12,
            "tau2": 0.17,
            "tau": 0.0,
            "sigma1": 0.30,
            "sigma2": 0.42,
            "sigma": 0.0,
            "m1": 5.0,
            "m2": 7.0,
        }
        c = self.coeff
        mref = (self.mag - c["m1"]) / (c["m2"] - c["m1"])

        #
        # Here we set the magnitude-dependent phi, tau, and sigma
        # to be used in the parent cless' getPGVandSTDDEVS()
        #
        if self.mag < c["m1"]:
            c["phi"] = c["phi1"]
            c["tau"] = c["tau1"]
            c["sigma"] = c["sigma1"]
        elif self.mag <= c["m2"]:
            c["phi"] = c["phi1"] + (c["phi2"] - c["phi2"]) * mref
            c["tau"] = c["tau1"] + (c["tau2"] - c["tau2"]) * mref
            c["sigma"] = c["sigma1"] + (c["sigma2"] - c["sigma2"]) * mref
        else:
            c["phi"] = c["phi2"]
            c["tau"] = c["tau2"]
            c["sigma"] = c["sigma2"]



import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from mpl_toolkits.mplot3d import axes3d


MAP_WIDTH = 10.0  # map width in degrees

# Temporary function for plotting maps.
# The idea is to separate plotting methods from modules like rupture.


def plot_rupture_wire3d(rupture, ax=None):
    """
    Method for making a simple representation of a Rupture instance.
    This method draws the outline of each quadrilateral in 3D.

    Args:
        rupture: A Rupture instance.
        ax: A matplotlib axis (optional).

    Returns:
        Matplotlib axis.
    """

    if ax is None:
        fig = plt.figure()
        ax = axes3d.Axes3D(fig)
        # ax = fig.add_subplot(111, projection="3d")
    else:
        if "zaxis" not in list(ax.properties().keys()):
            raise TypeError("Non-3d axes object passed to plot() method.")
    for quad in rupture.getQuadrilaterals():
        x = [p.longitude for p in quad]
        x.append(x[0])
        y = [p.latitude for p in quad]
        y.append(y[0])
        z = [-p.depth for p in quad]
        z.append(z[0])
        ax.plot(x, y, z, "k")
        ax.set_xlabel("Longitude")
        ax.set_ylabel("Latitude")
        ax.set_zlabel("Depth")

    return ax


def map_rupture(rupture):
    """
    Method for making a simple representation of a Rupture instance.
    This method draws the surface projection of the rupture on a map.

    Args:
        rupture: A Rupture instance.

    """
    rlats = rupture.lats
    rlons = rupture.lons

    # remove nans in these arrays
    rlats = rlats[~np.isnan(rlats)]
    rlons = rlons[~np.isnan(rlons)]

    minbufx = 0.2
    minbufy = 0.2
    lat1 = np.nanmin(rupture.lats)
    lat2 = np.nanmax(rupture.lats)
    dlat = lat2 - lat1
    bufy = np.max([dlat / 2, minbufy])
    lon1 = np.nanmin(rupture.lons)
    lon2 = np.nanmax(rupture.lons)
    dlon = lon2 - lon1
    bufx = np.max([dlon / 2, minbufx])

    ax = plt.axes(projection=ccrs.Robinson())
    bufx = MAP_WIDTH / 2.0
    bufy = MAP_WIDTH / 2.0
    ax.set_extent([lon1 - bufx, lon2 + bufx, lat1 - bufy, lat2 + bufy])
    ax.stock_img()
    ax.coastlines()
    ax.plot(rlons, rlats, "r", transform=ccrs.PlateCarree())
    return ax



# Local imports
from esi_shakelib.conversions.convert_imt import IMTConverter


class BommerAlarcon2006(IMTConverter):
    """
    Class for conversion between PGV (units of cm/s) and PSA05 (units of g)
    by Bommer and Alarcon (2006).

    - PSA05 stands for spectral acceleration with oscillator period of 0.5 sec
    - PGV is peak ground velocity.

    References:
        Bommer, J. J., & Alarcon, J. E. (2006). The prediction and use of peak
        ground velocity. Journal of Earthquake Engineering, 10(01), 1-31.
        `[link] <http://www.worldscientific.com/doi/abs/10.1142/S1363246906002463>`__
    """

    def __init__(self):
        super().__init__()
        # output_input dictionary where the key is the output
        # and the value is a list of the possible inputs
        self.output_input = {"PGV": ["PSA05"], "PSA05": ["PGV"]}
        self.conversion_factor = 1.0 / (20.0) * 100.0 * 9.81

    def convertAmps(self, imt_in, imt_out, imt):
        """
        Returns an array of converted IMT amplitude values.

        Args:
            imt_in (str): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            imt_out (str): OQ intensity measure type that the values will
                be converted to.
            imt (OpenQuake IMT): The intensity measurements of the input
                ground motions. Valid IMTs are PGV, and SA.

        Returns:
            array: Numpy array of amps converted from imt_in to imt_out.

        Raises:
            ValueError: If not a valid conversion.
        """
        # Verify that this is a valid conversion
        self._verifyConversion(imt_in, imt_out)
        imt_in = imt_in.upper().strip()
        imt_out = imt_out.upper().strip()
        conversion_factor = self.conversion_factor
        # Check which method to use
        if imt_in == "PSA05" and imt_out == "PGV":
            new_imt = self._convertToPGV(imt, conversion_factor)
        elif imt_in == "PGV" and imt_out == "PSA05":
            new_imt = self._convertToPSA05(imt, conversion_factor)
        else:
            raise ValueError(f"No conversion available from {imt_in!r} to {imt_out!r}")
        return new_imt

    @staticmethod
    def _convertToPGV(psa05, conversion_factor):
        """
        Convert PSA05 (spectral acceleration with oscillator period of 0.5 sec)
        in g to PGV cm/s.
        **Important:** PSA10 must be linear units.

        Args:
            psa05 (array): Numpy array or float of PSA05 values; linear units.

        Returns
            array: Numpy array or float of PGV converted from psa05.
        """
        return psa05 * conversion_factor

    @staticmethod
    def _convertToPSA05(pgv, conversion_factor):
        """
        Convert PGV in cm/s to PSA05 in g.
        **Important:** PGV must be linear units.

        Args:
            pgv (array): Numpy array or float of PGV values; linear units.

        Returns:
            array: Numpy array or float of PSA05 (spectral acceleration with
            oscillator period of 0.5 sec) converted from PGV.
        """
        return pgv / conversion_factor



# local imports
import re


def oq_to_file(oqimt):
    """Convert openquake IMT nomenclature to filename friendly form.

    Examples:
    SA(1.0) (Spectral Acceleration at 1 second) -> psa1p0
    SA(0.3) (Spectral Acceleration at 0.3 second) -> psa0p3
    SA(15.0) (Spectral Acceleration at 15 seconds) -> psa15p0
    SA(3) (Spectral Acceleration at 3 seconds) -> psa3p0
    SA(.5) (Spectral Acceleration at 0.5 seconds) -> psa0p5

    Args:
        oqimt (str): Openquake IMT nomenclature string.
    Returns:
        str: Filename friendly IMT string.
    Raises:
        ValueError: when there is no corresponding filename-friendly
            IMT representation.
    """
    if oqimt in ["PGA", "PGV", "MMI"]:
        return oqimt.lower()
    float_pattern = r"[-+]?\d*\.\d+|\d+"
    periods = re.findall(float_pattern, oqimt)
    if not len(periods):
        fmt = 'IMT string "%s" has no file-name friendly representation.'
        raise ValueError(fmt % oqimt)
    period = periods[0]
    if period.find(".") < 0:
        integer = period
        fraction = "0"
    else:
        integer, fraction = period.split(".")
        if not len(integer):
            integer = "0"
    fileimt = f"psa{integer}p{fraction}"
    return fileimt


def file_to_oq(fileimt):
    """Convert filename friendly IMT form to openquake form.

    Examples:
    psa1p0 (Spectral Acceleration at 1 second) -> SA(1.0)
    psa0p3 (Spectral Acceleration at 0.3 second) -> SA(0.3)
    psa15p0 (Spectral Acceleration at 15 seconds) -> SA(15.0)

    Args:
        fileimt (str): Filename friendly IMT string.
    Returns:
        str: Openquake IMT nomenclature string.
    """
    if fileimt in ["pga", "pgv", "mmi"]:
        return fileimt.upper()
    if "p" not in fileimt:
        fmt = "%s is not a valid filename-friendly IMT string."
        raise ValueError(fmt % fileimt)
    integer, fraction = fileimt.replace("psa", "").split("p")
    if not len(fraction):
        fmt = "%s is not a valid filename-friendly IMT string."
        raise ValueError(fmt % fileimt)
    oqimt = f"SA({integer}.{fraction})"
    return oqimt
